#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/5/24 14:37
@File    : parse_cfg.py
@Software: PyCharm
@Desc    : 
"""
import os
import warnings
from os import path as osp
from pathlib import Path
from typing import Any, Union

from .open_cfg import open_config
from .py_cfg import Config, ItemLazy

__all__ = ['auto_rslt_dir', 'parse_config', 'load_config']


def _check_emtpy_value(val: Any, memo: str='base.'):
    """Recursively detect empty val in dict"""
    warnings.warn("COM feature is deprecated", DeprecationWarning)
    if (not val) and not isinstance(val, (int, bool)):  # 若val被判否，且不是布尔（False）或int（0）类型，则是一个空Value。
        warnings.warn(f"{memo[:-1]} is a empty val: {val}")
    elif isinstance(val, dict):
        for k, v in val.items():
            _check_emtpy_value(v, memo + str(k) + '.')  # noqa
    else:
        pass


def auto_rslt_dir(file: str, config_root: str='./configs', trunc_cwd: bool=False) -> str:
    """根据配置文件的__file__属性，自动返回配置文件的rslt_dir。

    Args:
        file: 配置文件的__file__属性。
        config_root: 相对工作目录，路径之于配置目录者。
        trunc_cwd: 是否去掉当前工作目录。

    Returns:
        配置文件的rslt_dir，默认是__file__去掉配置目录和配置文件后的路径。
    """
    if not trunc_cwd:
        return osp.dirname(osp.relpath(file, config_root))
    else:
        return str(Path(file).relative_to(Path.cwd()).relative_to(config_root).parent)


def _process_yaml_config(config: dict, experiments_root: str, create_rslt_dir: bool=True):
    assert config.get('EXP_DIR') is None and config.get('TRAIN_DIR') is None and config.get('TEST_DIR') is None
    EXP_ID = config.get('EXP_ID')
    if EXP_ID is None:
        raise ValueError("CONFIG.EXP_ID must be defined")
    EXP_DIR = osp.join(experiments_root, str(EXP_ID))
    if create_rslt_dir:
        os.makedirs(EXP_DIR, exist_ok=True)
    config['EXP_DIR'] = EXP_DIR
    rslt_dir = EXP_DIR

    TRAIN_ID = config.get('TRAIN_ID')
    if TRAIN_ID is not None:
        TRAIN_DIR = osp.join(EXP_DIR, 'trains', str(TRAIN_ID))
        if create_rslt_dir:
            os.makedirs(TRAIN_DIR, exist_ok=True)
        rslt_dir = TRAIN_DIR
    else:
        TRAIN_DIR = EXP_DIR
    config['TRAIN_DIR'] = TRAIN_DIR

    TEST_ID = config.get('TEST_ID')
    if TEST_ID is not None:
        if TRAIN_ID is not None:
            TEST_DIR = osp.join(TRAIN_DIR, 'tests', str(TEST_ID))
        else:
            TEST_DIR = osp.join(EXP_DIR, 'tests', str(TEST_ID))
        if create_rslt_dir:
            os.makedirs(TEST_DIR, exist_ok=True)
        rslt_dir = TEST_DIR
    else:
        TEST_DIR = EXP_DIR
    config['TEST_DIR'] = TEST_DIR
    config['rslt_dir'] = rslt_dir
    if 'RAND_SEED' in config:
        config['rand_seed'] = config['RAND_SEED']
    return config


def _process_py_config(config: dict, config_path: str, experiments_root: str, config_root: str,
                       create_rslt_dir: bool=True):
    if isinstance(config, Config):
        # TODO 支持下级cfg删除上级cfg项。
        config.update_at_parser()
        # NOTE 1）此时归并COM，令后续IL计算、用户对COM无感。先update再归并，确保config能够接触、修改各层的所有COM树。
        # NOTE 2）归并可能令叶子间共享容器、IL函数（COM_会更新多个并形项）。故要确保
        # NOTE      a. 不直接修改叶子容器，而是覆盖叶子。
        # NOTE      b. IL函数无记忆，即多次调用，结果相同。
        # config.reduce_COM()

    if config.get('rslt_dir', ...) is ...:
        config['rslt_dir'] = auto_rslt_dir(config_path, config_root)

    if experiments_root != '':
        config['rslt_dir'] = osp.join(experiments_root, config['rslt_dir'])

    if create_rslt_dir:
        os.makedirs(config['rslt_dir'], exist_ok=True)

    return config


def parse_config(config_path: Union[str, dict], experiments_root: str=None, config_root: str='./configs',
                 create_rslt_dir: bool=True) -> Config:
    """Parse config from config path.

    This function will read yaml config from config path then create experiments dirs according to config file.

    Args:
        config_path: Path of yaml config.
        experiments_root: Root dictionary for experiments.
        config_root: Root dictionary of configs, for auto_rslt_dir only.
        create_rslt_dir: Whether to create rslt_dir.

    Returns:
        YAML config in Addict format
    """
    if create_rslt_dir and experiments_root is None:
        # NOTE 如果需要创建rslt_dir，那么experiments_root必须提供。
        # NOTE 相反，若之后experiments_root为None，肯定不需要创建rslt_dir，此时experiment_root只用于修正原来的rslt_dir。
        # NOTE 因此，不妨在这种情况下，将experiments_root视作''。
        raise ValueError("experiments_root must be provided when create_rslt_dir is True")

    # * Open config, get config dict.
    config, is_py = open_config(config_path, is_yaml=False)
    if config is None:
        raise RuntimeError(f"Failed to parse config at {config_path}")
    if isinstance(config_path, dict):
        # NOTE 当config_path是dict时，auto_rslt_dir不可为...——因为无法获取真正的config_path。
        # NOTE 此时，需要确保config['rslt_dir']已经被赋值。
        if not config.get('rslt_dir'):
            raise RuntimeError(f"config should indicate result save dir at "
                               f"config['rslt_dir'] = {config.get('rslt_dir')}")
        if config['rslt_dir'] is ...:
            raise RuntimeError(f"When config_path is dict, config['rslt_dir'] should not be ...")

    # * Create experiment dirs according to CONFIG
    if not is_py:
        config = _process_yaml_config(config, '' if experiments_root is None else experiments_root, create_rslt_dir)
    else:
        config = _process_py_config(config, config_path, '' if experiments_root is None else experiments_root,
                                    config_root=config_root,
                                    create_rslt_dir=create_rslt_dir)

    # * Check empty value and return
    # _check_emtpy_value(config, memo='cfg.')  # NOTE 似乎无此必要。
    return Config.from_dict(config)


def load_config(config_path: Union[str, dict], experiments_root: str=None, config_root: str='./configs',
                create_rslt_dir: bool=True) -> Config:
    """Parse config then compute lazy items and freeze.

    Args:
        config_path: See `parse_config`.
        experiments_root: See `parse_config`.
        config_root: See `parse_config`.
        create_rslt_dir: See `parse_config`.

    Returns:
        Frozen config with lazy items computed.
    """
    if isinstance(config_path, Config):
        config_path.unfreeze()  # 如果是load，接下来需要parse和compute_item_lazy，故解冻。反正最后总是会freeze。
    config = parse_config(config_path, experiments_root, config_root, create_rslt_dir)
    config = ItemLazy.compute_item_lazy(config)
    config.freeze()
    return config



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/5/29 16:38
@File    : file_md5.py
@Software: PyCharm
@Desc    : 
"""
from typing import Union
import hashlib
from os import PathLike

__all__ = ['file_md5']


def file_md5(file: Union[str, PathLike]) -> str:
    """Get md5 of file"""

    with open(file, 'rb') as f:
        # * read contents of the file.
        data = f.read()
        # * pipe contents of the file through.
        md5_ret = hashlib.md5(data).hexdigest()

    return md5_ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/7/23 3:45
@File    : validate.py
@Software: PyCharm
@Desc    : 
"""
from tqdm import tqdm
import numpy as np

import torch
from torch import nn
from torch.utils.tensorboard import SummaryWriter

from alchemy_cat.data import DataAuger
from alchemy_cat.contrib.voc import VOC_CLASSES
from alchemy_cat.contrib.metrics import SegmentationMetric
from alchemy_cat.alg import msc_flip_inference, find_nearest_odd_size

__all__ = ['validate_seg']


def validate_seg(model: nn.Module, val_auger: DataAuger, iteration: int, writer: SummaryWriter=None):
    """测试分割模型。

    Args:
        model: torch模型。
        val_auger: validation数据集的data_auger。
        writer: tensorboard的writer。
        iteration: 当前迭代次数。

    Returns:

    """
    print("\n================================== Validation ==================================")
    metric = SegmentationMetric(len(VOC_CLASSES), VOC_CLASSES)
    device = list(model.parameters())[0].device

    for _, img, label in tqdm(val_auger, total=len(val_auger),
                              desc="Validation Process", unit='samples', dynamic_ncols=True):
        bt_img = torch.from_numpy(img).to(device=device, dtype=torch.float32)[None, ...]
        score_map = msc_flip_inference(imgs=bt_img,
                                       model=model,
                                       msc_factors=[1.],
                                       is_flip=False,
                                       msc_aligner=lambda x: find_nearest_odd_size(x, min_n=3),
                                       softmax_norm=False
                                       ).cpu().numpy()[0]
        metric.update(np.argmax(score_map, axis=0), label)

    metric.print_statistics(importance=1)

    if writer is not None:
        writer.add_scalar('mIoU', metric.mIoU, iteration + 1)
        writer.add_scalar('Precision', metric.macro_avg_precision, iteration + 1)
        writer.add_scalar('Recall', metric.macro_avg_recall, iteration + 1)
        writer.add_scalar('Accuracy', metric.accuracy, iteration + 1)
    print("\n================================ Validation End ================================")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/11/25 20:51
@File    : __init__.py.py
@Software: PyCharm
@Desc    : 
"""
from .funcs import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/6/26 22:25
@File    : 转VOC上传.py
@Software: PyCharm
@Desc    : 
"""
import argparse
import os

import numpy as np
from PIL import Image
from tqdm import tqdm

parser = argparse.ArgumentParser()
parser.add_argument('-s', '--source', type=str)
parser.add_argument('-t', '--target', type=str)
args = parser.parse_args()

os.makedirs(args.target, exist_ok=True)

palette = []

for i in range(256):
    palette.extend((i, i, i))

palette[:3 * 21] = np.array([[0, 0, 0],
                             [128, 0, 0],
                             [0, 128, 0],
                             [128, 128, 0],
                             [0, 0, 128],
                             [128, 0, 128],
                             [0, 128, 128],
                             [128, 128, 128],
                             [64, 0, 0],
                             [192, 0, 0],
                             [64, 128, 0],
                             [192, 128, 0],
                             [64, 0, 128],
                             [192, 0, 128],
                             [64, 128, 128],
                             [192, 128, 128],
                             [0, 64, 0],
                             [128, 64, 0],
                             [0, 192, 0],
                             [128, 192, 0],
                             [0, 64, 128]], dtype='uint8').flatten()

for file in tqdm(os.listdir(args.source), desc="处理", unit="张", dynamic_ncols=True):
    im = Image.open(os.path.join(args.source, file))
    im = Image.fromarray(np.array(im).astype(np.uint8))
    im.putpalette(palette)
    im.save(os.path.join(args.target, file))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/8/10 17:50
@File    : voc_annotation.py
@Software: PyCharm
@Desc    :
"""
from typing import Tuple, Optional, List, Dict, Any

from collections import defaultdict
from xml.etree import ElementTree as ET
from xml.dom import minidom

__all__ = ['dump_voc_annotation', 'load_voc_annotation', 'count_obj_num']


def _sub_element_with_text(parent: ET.Element, tag: str, text: Optional[str]=None) -> ET.Element:
    tmp = ET.SubElement(parent, tag)
    if text is not None:
        tmp.text = text
    return tmp


def dump_voc_annotation(xml_file: str,
                        folder: Optional[str]=None,
                        filename: Optional[str]=None,
                        source_database: Optional[str]=None, source_annotation: Optional[str]=None,
                        source_image: Optional[str]=None,
                        size: Optional[Tuple[int, int]]=None,
                        segmented: Optional[int]=None,
                        objects: Optional[List[Dict[str, Any]]]=None) -> str:
    # * Create <annotation>
    root = ET.Element("annotation")

    # * Create <folder>
    if folder is not None:
        _sub_element_with_text(root, 'folder', folder)

    # * Create <filename>
    if filename is not None:
        _sub_element_with_text(root, 'filename', filename)

    # * Create <source>
    if source_database is not None:
        assert source_annotation is not None
        assert source_image is not None
        source_tag = _sub_element_with_text(root, 'source')
        _sub_element_with_text(source_tag, 'database', source_database)
        _sub_element_with_text(source_tag, 'annotation', source_annotation)
        _sub_element_with_text(source_tag, 'image', source_image)

    # * Create <size>
    if size is not None:
        img_h, img_w = size
        size_tag = _sub_element_with_text(root, 'size')
        _sub_element_with_text(size_tag, 'width', str(img_w))
        _sub_element_with_text(size_tag, 'height', str(img_h))
        _sub_element_with_text(size_tag, 'depth', '3')

    # * Create <segmented>
    if segmented is not None:
        _sub_element_with_text(root, 'segmented', str(segmented))

    # * Create <object>
    if objects is not None:
        for obj in objects:
            # * Get object attrib
            name = obj['name']
            pose = obj.get('pose', 'Unspecified')
            truncated = obj.get('truncated', '0')
            difficult = obj.get('difficult', '0')
            xmin, ymin, xmax, ymax = obj['bndbox']

            # * Create object tag
            object_tag = _sub_element_with_text(root, 'object')
            _sub_element_with_text(object_tag, 'name', name)
            _sub_element_with_text(object_tag, 'pose', pose)
            _sub_element_with_text(object_tag, 'truncated', str(truncated))
            _sub_element_with_text(object_tag, 'difficult', str(difficult))
            bndbox_tag = _sub_element_with_text(object_tag, 'bndbox')
            _sub_element_with_text(bndbox_tag, 'xmin', str(xmin))
            _sub_element_with_text(bndbox_tag, 'ymin', str(ymin))
            _sub_element_with_text(bndbox_tag, 'xmax', str(xmax))
            _sub_element_with_text(bndbox_tag, 'ymax', str(ymax))

    xml_str = minidom.parseString(ET.tostring(root)).childNodes[0].toprettyxml(indent='	')  # No xml header

    with open(xml_file, 'w') as f:
        f.write(xml_str)

    return xml_str


def load_voc_annotation(xml_file: str) -> Dict[str, Any]:
    def to_int(obj):
        return int(obj) if obj is not None else obj

    tree = ET.parse(xml_file)
    root = tree.getroot()

    ret = {}

    # * Load <folder>
    ret['folder'] = root.findtext('folder')

    # * Load <filename>
    ret['filename'] = root.findtext('filename')

    # * Load <source>
    ret['source_database'] = root.findtext('source/database')
    ret['source_annotation'] = root.findtext('source/annotation')
    ret['source_image'] = root.findtext('source/image')

    # * Load <size>
    ret['size'] = (to_int(root.findtext('size/height')), to_int(root.findtext('size/width')))

    # * Load <segmented>
    ret['segmented'] = to_int(root.findtext("segmented"))

    # * Load <object>
    ret['objects'] = []
    for object_tag in root.findall('object'):
        ret['objects'].append({
            'name': object_tag.findtext('name'),
            'pose': object_tag.findtext('pose'),
            'truncated': to_int(object_tag.findtext('truncated')),
            'difficult': to_int(object_tag.findtext('difficult')),
            'bndbox': [to_int(object_tag.findtext('bndbox/xmin')),
                       to_int(object_tag.findtext('bndbox/ymin')),
                       to_int(object_tag.findtext('bndbox/xmax')),
                       to_int(object_tag.findtext('bndbox/ymax'))]
        })

    return ret


def count_obj_num(anno: Dict[str, Any]) -> Dict[str, int]:
    obj_num = defaultdict(int)

    for obj in anno['objects']:
        obj_num[obj['name']] += 1

    return dict(obj_num)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/2/4 22:12
@File    : scheduler_chainer.py
@Software: PyCharm
@Desc    : 
"""
import bisect

from torch.optim.lr_scheduler import _LRScheduler

__all__ = ['SchedulerChainer']


class SchedulerChainer(object):

    def __init__(self, schedulers: list[_LRScheduler], milestones: list[int | float]):
        """拼接多个调率器。

        Args:
            schedulers: 调率器列表。
            milestones: 调率器切换里程碑。
        """
        assert len(schedulers) > 0
        assert len(milestones) == len(schedulers) - 1

        self.schedulers = schedulers
        self.milestones = milestones + [float('inf')]

        self._iteration = -1

    @property
    def iteration(self) -> int:
        """返回当前迭代号（从0开始）。"""
        return self._iteration

    def step(self, iteration: int | None=None):
        if iteration is not None:
            self._iteration = iteration
        else:
            self._iteration += 1
        todo_iter = self.iteration + 1  # 调率器的作用迭代。

        if len(self.schedulers) == 1:  # 只有一个scheduler，迭代到底。
            self.schedulers[0].step()
        else:
            right_milestone_idx = bisect.bisect_left(self.milestones, todo_iter)
            self.schedulers[right_milestone_idx].step()
            if todo_iter == self.milestones[right_milestone_idx]:
                self.schedulers[right_milestone_idx + 1].step()



default_scope = 'mmseg'
env_cfg = dict(
    cudnn_benchmark=False,
    mp_cfg=dict(mp_start_method='fork', opencv_num_threads=0),
    dist_cfg=dict(backend='nccl'),
)
vis_backends = [dict(type='LocalVisBackend'),
                dict(type='TensorboardVisBackend')]
visualizer = dict(
    type='SegLocalVisualizer', vis_backends=vis_backends, name='visualizer')
log_processor = dict(by_epoch=False)
log_level = 'INFO'
load_from = None
resume = False

tta_model = dict(type='SegTTAModel')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/2/22 20:19
@File    : voc上色.py
@Software: PyCharm
@Desc    : 
"""
import argparse
import multiprocessing as mp
import os
import os.path as osp

import numpy as np
from PIL import Image
from alchemy_cat.contrib.coco import COCO
from alchemy_cat.contrib.evaluation.semantics_segmentation import eval_preds
from alchemy_cat.data.plugins import arr2PIL
from tqdm import tqdm


def 上色_label_file(label_file: str):
    if not label_file.endswith('.png'):
        return

    label = np.array(Image.open(osp.join(args.source, label_file)), dtype=np.uint8)
    color_label = dt.label_map2color_map(label)
    arr2PIL(color_label, order='RGB').save(osp.join(args.target, label_file))


def ignore2bg(lb: np.ndarray):
    lb[lb == 255] = 0
    return lb


parser = argparse.ArgumentParser()
parser.add_argument('-e', '--eval', type=int, default=0)
parser.add_argument('-c', '--colorize', type=int, default=0)
parser.add_argument('-s', '--source', type=str)
parser.add_argument('-t', '--target', type=str)
parser.add_argument('--label_type', type=str, default='')
parser.add_argument('--split', type=str, default='train')
args = parser.parse_args()

dt = COCO('datasets', split=args.split, label_type=args.label_type, cls_labels_type=None)

if args.colorize:
    os.makedirs(args.target, exist_ok=True)

    with mp.Pool(int(mp.cpu_count() * 0.2)) as p:
        for _ in tqdm(p.imap_unordered(上色_label_file, files := os.listdir(args.source), chunksize=10),
                      dynamic_ncols=True, desc="处理", unit="张", total=len(files)):
            pass

if args.eval:
    metric = eval_preds(class_num=dt.class_num,
                        class_names=dt.class_names,
                        preds_dir=str(args.source),
                        preds_ignore_label=dt.ignore_label,
                        gts_dir=dt.label_dir,
                        gts_ignore_label=dt.ignore_label,
                        pred_preprocess=ignore2bg,
                        result_dir=None,
                        importance=0,
                        eval_individually=False,
                        take_pred_ignore_as_a_cls=False)

    metric.print_statistics(0)



""" class abstracting data passed as inputs and saved as outputs """
from copy import deepcopy

from alchemy_cat.dag.errors import PyungoError
from alchemy_cat.dag.io import Output


class Data:
    def __init__(self, inputs, slim=False):
        self._slim = slim

        if self._slim:
            self._inputs = inputs
        else:
            self._inputs = deepcopy(inputs)

        self._outputs = {}

    @property
    def inputs(self):
        return self._inputs

    @property
    def outputs(self):
        return self._outputs

    def __getitem__(self, map):
        try:
            return self._inputs[map]
        except KeyError:
            return self._outputs[map]

    def __setitem__(self, map, output_value):
        self._outputs[map] = output_value

    def __contains__(self, map):
        return (map in self._inputs) or (map in self._outputs)

    def check_inputs(self, dag_nec_input_maps: set, dag_opt_input_maps: set, dag_output_names: set):
        """ Check input is legal. """
        # inputs in data can't have the same name of any the output
        data_input_maps = set(self.inputs.keys())
        dag_nec_input_maps = dag_nec_input_maps
        dag_opt_input_maps = dag_opt_input_maps
        dag_output_names = dag_output_names

        diff = data_input_maps & dag_output_names
        if diff:
            msg = 'The following inputs are already used in the model: {}'
            raise PyungoError(msg.format(sorted(list(diff))))
        # inputs in data should be able to provide inputs needed for calculate
        diff = dag_nec_input_maps - dag_output_names - data_input_maps
        if diff:
            msg = 'The following inputs are needed: {}'.format(sorted(list(diff)))
            raise PyungoError(msg)
        # All inputs should be used in the calculation
        diff = data_input_maps - dag_nec_input_maps - dag_opt_input_maps
        if diff:
            msg = 'The following inputs are not used by the model: {}'
            raise PyungoError(msg.format(sorted(list(diff))))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/3/9 21:49
@File    : inf_loader.py
@Software: PyCharm
@Desc    : 
"""
from torch.utils.data import DataLoader

__all__ = ['inf_loader']


def inf_loader(loader: DataLoader):
    """接受一个DataLoader对象作为参数，每次迭代都会返回下一个batch数据。

    Args:
        loader: 数据加载器。

    Returns:
        dataloader的下一个batch。
    """
    epoch_iter = iter(loader)
    while True:
        try:
            bt = next(epoch_iter)
        except StopIteration:
            epoch_iter = iter(loader)
            bt = next(epoch_iter)
        yield bt



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/2/19 17:25
@File    : tag_exps.py
@Software: PyCharm
@Desc    : 根据git提交中的配置文件，给指定提交打上标签，名之以实验标识。
"""
import os
import os.path as osp
import argparse
import subprocess

from alchemy_cat.py_tools import gprint, yprint

"""
 六言·参数含义
   乾隆再世
 在指定仓库下，
 列出提交更改。
 在配置目录下，
 找到配置文件。
 文件对应实验，
 标到指定提交。
"""
parser = argparse.ArgumentParser()
parser.add_argument('-r', '--repo', default='.', help="Git repo root dir. ", type=str)
parser.add_argument('-s', '--show', default='HEAD', help="Commit to show file changed. ", type=str)
parser.add_argument('-c', '--config_dir', default='configs', help="Relative path to repo dir of config dir. ", type=str)
parser.add_argument('-n', '--config_name', default='cfg.py', help="Config files' name. ", type=str)
parser.add_argument('-a', '--at', default='HEAD', help="Tag experiment ID at which commit. ", type=str)
parser.add_argument('-k', '--skip_tagged', default=0, help="If true, auto skip exp already tagged. ", type=int)
args = parser.parse_args()


def removeprefix(text: str, prefix: str) -> str:
    """Alternative for str.removeprefix in python 3.9.

    See Also:
        https://docs.python.org/3/library/stdtypes.html#str.removeprefix
    """
    if text.startswith(prefix):
        return text[len(prefix):]
    return text


# * 在指定仓库下，列出某提交更改的文件。
p = subprocess.run(['git', 'show', '--pretty=', '--name-only', args.show],
                   cwd=osp.normpath(args.repo),
                   capture_output=True, encoding='utf-8', check=False)
if p.stderr:
    raise RuntimeError(f"Failed to show filed changed at commit {args.show}, git command returns: \n"
                       f"{p.stderr}")

changed_files = [osp.normpath(file) for file in p.stdout.splitlines()]

print(f"In repo {osp.abspath(args.repo)}, find following changed files at commit {args.show}: \n" +
      '\n'.join(changed_files), end="\n\n")

# * 基于更改文件，在配置目录下，找到配置文件。
config_dir = osp.normpath(args.config_dir) + os.sep
config_files = [file for file in changed_files if file.startswith(config_dir)] \
    if config_dir != '.' + os.sep else changed_files
config_files = [file for file in config_files if osp.basename(file) == args.config_name]

print(f"Find following config files in config dir {config_dir} named {args.config_name}: \n" +
      "\n".join(config_files), end="\n\n")

# * 基于配置文件，生成实验标识。
exp_ids = []
for file in config_files:
    file_wo_prefix = removeprefix(file, config_dir) if config_dir != '.' + os.sep else file
    exp_dir = osp.dirname(file_wo_prefix)
    if not exp_dir:
        raise RuntimeError(f"Config file {file} is not in exp dir. ")
    exp_ids.append(exp_dir.replace(os.sep, '-'))

# * 获取仓库已有标签。
p = subprocess.run(['git', 'tag'],
                   cwd=osp.normpath(args.repo),
                   capture_output=True, encoding='utf-8', check=False)
if p.stderr:
    raise RuntimeError(f"Failed to list tags at repo {osp.abspath(args.repo)}, git command returns: \n"
                       f"{p.stderr}")

tags = p.stdout.splitlines()

# * 对找到的每个配置文件及对应实验标识，由用户决定，是否要在指定提交打标签。
for i, (config_file, exp_id) in enumerate(zip(config_files, exp_ids)):
    if args.skip_tagged and (exp_id in tags):
        yprint(f"[{i+1}/{len(config_files)}] Skip {exp_id} for {config_file}, for it was already tagged. ", end="\n\n")
        continue

    while True:
        selection = input(f"For {config_file}, tag {exp_id} at commit {args.at}?[y/n]: ")
        if selection == 'y':
            p = subprocess.run(['git', 'tag', exp_id, args.at],
                               cwd=osp.normpath(args.repo),
                               capture_output=True, encoding='utf-8', check=False)
            if p.stderr:
                raise RuntimeError(f"Failed to tag {exp_id} at commit {args.at}, git command returns: \n"
                                   f"{p.stderr}")
            gprint(f"[{i+1}/{len(config_files)}] Tag {exp_id} at commit {args.at}. ", end="\n\n")
            break
        elif selection == 'n':
            if exp_id not in tags:
                yprint(f"[{i+1}/{len(config_files)}] Don't tag {exp_id} at commit {args.at}, while "
                       f"this id is still not tagged. ", end="\n\n")
            else:
                yprint(f"[{i+1}/{len(config_files)}] Don't tag {exp_id} at commit {args.at}, attention "
                       f"this id is already tagged. Use -k 1 to avoid re-tag. ", end="\n\n")
            break
        else:
            print(f"Unknown selection {selection}, please re-enter. ")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/11/8 21:26
@File    : split_set.py
@Software: PyCharm
@Desc    : 
"""
import argparse
import os
import os.path as osp
import glob
from math import ceil
import random

parser = argparse.ArgumentParser()
parser.add_argument('-r', '--root', type=str)
parser.add_argument('-t', '--train_ratio', default=0.75, type=float)
parser.add_argument('-v', '--val_ratio', default=0.25, type=float)
args = parser.parse_args()

# * Rand Seed
kSeed = 0

# * Dir & files
kAnnoDir = osp.join(args.root, 'Annotations')

kTrainFile = osp.join(args.root, 'ImageSets', 'Main', 'train.txt')
kValFile = osp.join(args.root, 'ImageSets', 'Main', 'val.txt')
kTestFile = osp.join(args.root, 'ImageSets', 'Main', 'test.txt')
kTrainValFile = osp.join(args.root, 'ImageSets', 'Main', 'trainval.txt')

os.makedirs(osp.join(args.root, 'ImageSets', 'Main'), exist_ok=True)

# * Set Rand Seed
random.seed(kSeed)

# * Get names
names = [osp.splitext(osp.basename(file))[0] for file in glob.glob(osp.join(kAnnoDir, '*.xml'))]

# * Compute split size
train_num = ceil(len(names) * args.train_ratio)
assert train_num <= len(names)

val_num = ceil(len(names) * args.val_ratio)
if train_num + val_num > len(names):
    val_num = len(names) - train_num

test_num = len(names) - train_num - val_num

# * Shuffle names & get splits
random.shuffle(names)
train_names = names[:train_num]
val_names = names[train_num:train_num + val_num]
test_names = names[train_num + val_num:train_num + val_num + test_num]
trainval_names = train_names + val_names

# * Write names
with open(kTrainFile, 'w') as f:
    f.write('\n'.join(train_names) + '\n')

with open(kValFile, 'w') as f:
    f.write('\n'.join(val_names) + '\n')

with open(kTestFile, 'w') as f:
    f.write('\n'.join(test_names) + '\n')

with open(kTrainValFile, 'w') as f:
    f.write('\n'.join(trainval_names) + '\n')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/5/25 21:38
@File    : cfg.py
@Software: PyCharm
@Desc    : 
"""
try:
    from addict import Dict
except ImportError:
    Dict = dict

from alchemy_cat.py_tools import Config


cfg = config = Config()

# * 二级子树。
cfg.foo0.foo1.a = 1
cfg.foo0.foo1.b = {'a': 2, 'b': 3}
# * 一级子树。
cfg.foo2.fix = 4
cfg.foo2.a_max = Dict({'c': 5, 'd': 6})
# * 叶子。
cfg.c = 7
cfg.d = Dict({'e': 8, 'f': 9})



# -- tune_train.py --
import argparse, json, os, subprocess, torch, sys
from alchemy_cat.dl_config import Config, Cfg2TuneRunner

parser = argparse.ArgumentParser(description='Tuning AlchemyCat MNIST Example')
parser.add_argument('-c', '--cfg2tune', type=str)
args = parser.parse_args()

# Set `pool_size` to GPU num, will run `pool_size` of configs in parallel
runner = Cfg2TuneRunner(args.cfg2tune, experiment_root='/tmp/experiment', pool_size=torch.cuda.device_count())

@runner.register_work_fn  # How to run config
def work(pkl_idx: int, cfg: Config, cfg_pkl: str, cfg_rslt_dir: str) -> ...:
    subprocess.run([sys.executable, 'train.py', '-c', cfg_pkl],
                   env=os.environ | {'CUDA_VISIBLE_DEVICE': f'pkl_idx % torch.cuda.device_count()'})

@runner.register_gather_metric_fn  # How to gather metric for summary
def gather_metric(cfg: Config, cfg_rslt_dir: str, run_rslt: ..., param_comb: dict[str, tuple[..., str]]) -> dict[str, ...]:
    return json.load(open(os.path.join(cfg_rslt_dir, 'eval.json')))

runner.tuning()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: __init__.py.py
@time: 2020/2/26 23:15
@desc:
"""
from alchemy_cat.contrib.metrics.conf_matrix_bases_metrics import *
from alchemy_cat.contrib.metrics.superpixel_metrics import *
from alchemy_cat.contrib.metrics.seg_mask_metric import *



# -- configs/mnist/plain_usage/cfg.py --

import torch.optim.lr_scheduler as sched
import torchvision.models as model
import torchvision.transforms as T
from torch import optim
from torchvision.datasets import MNIST

from alchemy_cat.dl_config import Config

cfg = Config()

cfg.rand_seed = 0

# -* Set datasets.
cfg.dt.cls = MNIST
cfg.dt.ini.root = '/tmp/data'
cfg.dt.ini.transform = T.Compose([T.Grayscale(3), T.ToTensor(), T.Normalize((0.1307,), (0.3081,)),])

# -* Set data loader.
cfg.loader.ini.batch_size = 128
cfg.loader.ini.num_workers = 2

# -* Set model.
cfg.model.cls = model.resnet18
cfg.model.ini.num_classes = 10

# -* Set optimizer.
cfg.opt.cls = optim.AdamW
cfg.opt.ini.lr = 0.01

# -* Set scheduler.
cfg.sched.epochs = 30
cfg.sched.warm_epochs = 3

cfg.sched.warm.cls = sched.LinearLR
cfg.sched.warm.ini.total_iters = 3
cfg.sched.warm.ini.start_factor = 1e-05
cfg.sched.warm.ini.end_factor = 1.0

cfg.sched.main.cls = sched.CosineAnnealingLR
cfg.sched.main.ini.T_max = 27

# -* Set logger.
cfg.log.save_interval = 6



import ast
import inspect

from alchemy_cat.dag.errors import PyungoError


def get_function_return_names(fct):
    """ Return variable name(s) or return statement of the given function """
    lines = inspect.getsourcelines(fct)
    outputs = None
    for line in lines[0][::-1]:
        stripped = line.strip()
        if 'def' in stripped:
            # NOTE: This work as def is a reserved keyword which will trigger
            # invalid syntax if misused
            msg = 'No return statement found in {}'
            raise PyungoError(msg.format(fct.__name__))
        ast_tree = ast.parse(stripped)
        for ast_node in ast.walk(ast_tree):
            if isinstance(ast_node, ast.Return):
                if isinstance(ast_node.value, ast.Name):
                    outputs = [ast_node.value.id]
                elif isinstance(ast_node.value, ast.Tuple):
                    outputs = [
                        elt.id for elt in ast_node.value.elts
                        if isinstance(elt, ast.Name)
                    ]
                else:
                    name = ast_node.value.__class__.__name__
                    msg = ('Variable name or Tuple of variable names are '
                           'expected, got {}'.format(name))
                    raise PyungoError(msg)
                break
        if outputs:
            break
    return outputs


def run_node(node):
    """Python multiprocessing works strangely in windows. The pool function needed to be
    defined globally

    Args:
        node (Node): Node to be called

    Returns:
        rslts: Node's call output
    """
    return node.run_with_loaded_inputs()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/5/8 22:18
@File    : viz_cam.py
@Software: PyCharm
@Desc    : 
"""
from typing import List

import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np

from alchemy_cat.contrib.voc import label_map2color_map

from alchemy_cat.acplot.subplots_row_col import square

__all__ = ["viz_cam"]


def viz_cam(fig: plt.Figure,
            img_id: str, img: np.ndarray, label: np.ndarray,
            cls_in_label: np.ndarray, cam: np.ndarray, cls_names: List[str],
            gamma: float = 1.0, blend_alpha: float = 0.5,
            get_row_col: callable=square, color_map: callable=label_map2color_map) -> plt.Figure:
    """ 可视化CAM。作图结果为：图像+label，图像+CAM0，图像+CAM1，······。

    Args:
        fig: plt的Figure对象。
        img_id: 图像的编号。
        img: 图像数组。
        label: 标签数组。
        cls_in_label: 长度为类别数的0,1数组，表示那些类别为正类。正类名字会被标红。
        cam: 类别激活图数组。
        cls_names: 类别名字。
        gamma: 若不为1，则对CAM做gamma矫正，但此时无法画出色带。
        blend_alpha: 叠加图像时的透明度。
        get_row_col: 函数，返回作图的行列数。
        color_map: 函数，输入单通道的数值标签，输出3通道的彩色标签。

    Returns:
        plt的Figure对象。
    """
    assert cam.shape[1:] == img.shape[:2]
    assert cam.shape[0] + 1 == len(cls_names)
    assert len(cls_names) == len(cls_in_label)

    # * Get grid size
    row_num, col_num = get_row_col(len(cls_names))

    # * Plot ref img
    ax: plt.Axes = fig.add_subplot(row_num, col_num, 1)

    ax.imshow(img)
    ax.imshow(color_map(label.copy()), alpha=blend_alpha)
    ax.set_title(img_id, fontsize='smaller')
    ax.axis("off")

    # * Show CAM
    for fore_cls, cls_name in enumerate(cls_names[1:]):
        # * Show CAM with bottom img.
        ax = fig.add_subplot(row_num, col_num, fore_cls + 2)
        ax.imshow(img)

        # * Get CAM
        cls_cam = cam[fore_cls, ...].astype(np.float32)

        if gamma == 1.0:
            mappable = ax.imshow(cls_cam, cmap=plt.get_cmap('jet'), alpha=blend_alpha)
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="2.5%", pad=0.05)
            cax.tick_params(labelsize=3, length=2, width=0.5, pad=0.5)

            fig.colorbar(mappable=mappable, cax=cax)
        else:
            cls_viz_cam = ((cls_cam - cls_cam.min()) / (cls_cam.max() - cls_cam.min())) ** gamma
            ax.imshow(cls_viz_cam, cmap=plt.get_cmap('jet'), alpha=blend_alpha)

        # * Set CAM Title
        font_dict = dict(color='red', fontweight='bold') if cls_in_label[fore_cls + 1] else dict()
        ax.set_title(cls_name, fontsize='smaller', **font_dict)
        ax.axis("off")

    fig.tight_layout(pad=0.2, h_pad=0.1, w_pad=0.2)
    return fig



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/6/22 15:07
@File    : voc_cooccur_analysis.py
@Software: PyCharm
@Desc    : 
"""
from typing import Optional

import numpy as np

from alchemy_cat.contrib.voc import VOCAug, VOC_CLASSES
from alchemy_cat.py_tools.type import tolist

kVOCSplit = 'train_aug'
kForeClsNum = len(VOC_CLASSES[1:])


def cls_in_label2tag(cls_in_label: np.ndarray):
    return np.nonzero(cls_in_label)[0]


def label2cls_in_label(label: np.ndarray, cls_num, ignore_label) -> np.ndarray:
    return (np.bincount(label.ravel(), minlength=ignore_label + 1) != 0).astype(np.int64)[:cls_num]


def get_permutations(elem_set: list, permutation_len: int, current_permutation: Optional[list]=None):
    assert permutation_len > 0
    assert elem_set

    if current_permutation is None:
        current_permutation = []

    for idx, elem in enumerate(elem_set):
        appended_permutation = current_permutation + [elem]
        if len(appended_permutation) == permutation_len:
            yield appended_permutation
        else:
            retain_set = elem_set[:]
            retain_set.pop(idx)
            yield from get_permutations(retain_set, permutation_len, appended_permutation)


def tag2cooccur_idx(tag: np.ndarray):
    """Get cooccur_idx from tag. e.g. tag = [1], cooccur_idx = (1, 1). tag = [0, 1, 5],
    cooccur_idx = ([0, 0, 1, 1, 5, 5], [1, 5, 0, 5, 0, 1]).

    Args:
        tag: img label tag.

    Returns:
        cooccur_idx
    """
    if len(tag) == 1:
        return tag[0], tag[0]
    else:
        return tuple(list(hw) for hw in zip(*get_permutations(tolist(tag), 2)))


if __name__ == "__main__":
    cooccur_matrix = np.zeros((kForeClsNum, kForeClsNum), dtype=np.int64)

    dt = VOCAug(split=kVOCSplit)

    for _, _, label in dt:
        fore_in_label = label2cls_in_label(label, kForeClsNum + 1, VOCAug.ignore_label)[1:]
        tag = cls_in_label2tag(fore_in_label)
        cooccur_idx = tag2cooccur_idx(tag)
        cooccur_matrix[cooccur_idx] += 1

    print(cooccur_matrix)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/5/6 22:09
@File    : cfg.py.py
@Software: PyCharm
@Desc    : 
"""
from alchemy_cat.py_tools import Cfg2Tune, Param2Tune

cfg = config = Cfg2Tune()

cfg.rslt_dir = 'run_time_mutual_depend_config'

# * Param on same Cfg2Tune
cfg.foo0.foo1.a = Param2Tune([0, 1, 2])


def b_subject_to(_):
    if cfg.foo0.foo1.a.cur_val == 0:
        return True

    return cfg.c.cur_val


cfg.foo0.foo1.b = Param2Tune([.1, .2], b_subject_to)
# * Param on root Cfg2Tune
cfg.c = Param2Tune([False, True], subject_to=lambda x: x == (10 * cfg.foo0.foo1.b.cur_val > cfg.foo0.foo1.a.cur_val))
# * Static Value
cfg.foo0.a_max = 2



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/2/22 20:19
@File    : colorize_voc.py
@Software: PyCharm
@Desc    : 
"""
import typing as t

import argparse
import multiprocessing as mp
import os
import os.path as osp
from functools import partial

import numpy as np
from PIL import Image
from alchemy_cat.contrib.evaluation.semantics_segmentation import eval_preds
from alchemy_cat.contrib.voc import VOCAug2
from alchemy_cat.contrib.voc import label_map2color_map
from alchemy_cat.data.plugins import arr2PIL
from tqdm import tqdm


def colorize_label_file(label_file: str, source: str, target: str, l2c: t.Callable[[np.ndarray], np.ndarray]=None):
    if not label_file.endswith('.png'):
        return

    if l2c is None:
        l2c = label_map2color_map

    label = np.array(Image.open(osp.join(source, label_file)))
    color_label = l2c(label).astype(np.uint8)
    arr2PIL(color_label, order='RGB').save(osp.join(target, label_file))


def ignore2bg(lb: np.ndarray):
    lb[lb == 255] = 0
    return lb


def colorize_voc(source: str, target: str, num_workers: int=0, is_eval: bool=False,
                 l2c: t.Callable[[np.ndarray], np.ndarray]=None):
    os.makedirs(target, exist_ok=True)

    label_files = os.listdir(source)
    worker = partial(colorize_label_file, source=source, target=target, l2c=l2c)

    if num_workers > 0:
        with mp.Pool(num_workers) as p:
            for _ in tqdm(p.imap_unordered(worker, label_files, chunksize=10), total=len(label_files),
                          dynamic_ncols=True, desc="上色VOC", unit="张", miniters=len(label_files)//10):
                pass
    else:
        for _ in tqdm(map(worker, label_files), total=len(label_files),
                      dynamic_ncols=True, desc="上色VOC", unit="张", miniters=len(label_files)//10):
            pass

    if is_eval:
        dt = VOCAug2('datasets')

        metric = eval_preds(class_num=dt.class_num,
                            class_names=dt.class_names,
                            preds_dir=str(source),
                            preds_ignore_label=dt.ignore_label,
                            gts_dir=dt.label_dir,
                            gts_ignore_label=dt.ignore_label,
                            pred_preprocess=ignore2bg,
                            result_dir=None,
                            importance=0,
                            eval_individually=False,
                            take_pred_ignore_as_a_cls=False)

        metric.print_statistics(0)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-e', '--is_eval', type=int, default=0)
    parser.add_argument('-n', '--num_workers', type=int, default=1)
    parser.add_argument('-s', '--source', type=str)
    parser.add_argument('-t', '--target', type=str)
    _args = parser.parse_args()
    colorize_voc(source=_args.source, target=_args.target, num_workers=_args.num_workers, is_eval=bool(_args.is_eval))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: voc2012_auger.py
@time: 2020/3/21 13:41
@desc:
"""
from typing import Union, Iterable

import numpy as np
import matplotlib.pyplot as plt

from alchemy_cat.data import DataAuger
from alchemy_cat.acplot import HWC2CHW, CHW2HWC
from alchemy_cat.data.plugins.augers import RandColorJitter, RandMirror, RandScale, RandCrop, pad_img_label, \
    int_img2float32_img, MultiMirror, centralize
from alchemy_cat.acplot.figure_wall import RowFigureWall, ColumnFigureWall
from alchemy_cat.contrib.voc.voc2012seg import VOC, VOCAug, label_map2color_map
from alchemy_cat.contrib.voc.utils import VOC_CLASSES

__all__ = ['VOCTrainAuger', 'VOCClsTrainAuger', 'VOCTestAuger', 'VOCClsTestAuger', 'attach_cls', 'collect_example',
           'lb2cls_lb']


def lb2cls_lb(label: np.ndarray) -> np.ndarray:
    return (np.bincount(label.ravel(), minlength=VOC.ignore_label + 1) != 0).astype(np.uint8)[:len(VOC_CLASSES)]


def attach_cls(example):
    img_id, img, label = example
    cls_in_label = lb2cls_lb(label)
    return img_id, img, label, cls_in_label


def collect_example(img_id, img, label):
    return img_id, img.copy(), label.copy() # Given continuous arr


class _VOCBaseAuger(DataAuger):
    def build_graph(self):
        @self.graph.register(inputs=['example'], outputs=['img_id', 'img', 'label'])
        def split_example(example):
            return example


class VOCTrainAuger(_VOCBaseAuger):
    """DataAuger for VOC Dataset when training

    The input img can be augered by random color jittering, random scaling, random crop and mirroring.
    """

    def __init__(self, dataset: Union[VOC, VOCAug],
                 verbosity: int=0, pool_size: int=0, slim: bool=False, rand_seed_log: str=None,
                 is_multi_mirror: bool=False, is_color_jitter: bool=False,
                 scale_factors: Iterable=(1.0), crop_size: Union[int, Iterable[int]]=321):
        """DataAuger for VOC Dataset when training

        Args:
            dataset: VOC or VOCAug dataset
            verbosity: See alchemy_cat.data.DataAuger
            pool_size: See alchemy_cat.data.DataAuger
            slim: See alchemy_cat.data.DataAuger
            rand_seed_log: See alchemy_cat.data.DataAuger
            is_multi_mirror: If False, img and label will be random mirrored, else both mirrored and not mirrored
                img/label will be given in augmentation.(Default: False)
            is_color_jitter: If True, use color jitter the input img. (Default: False)
            scale_factors: Scale factors for random scale. eg. [0.5, 1, 1.5] means the img (and label) will be
                rand scale with factor 0.5, 1.0 or 1.5(Default: (1.0))
            crop_size: Crop size. If size is int, the crop_height=value, crop_width=value. Else will be parsed as
                crop_height=list(value)[0], crop_width=list(value)[1]

        See Also:
            alchemy_cat.data.DataAuger
        """
        self.is_multi_mirror = is_multi_mirror
        self.is_color_jitter = is_color_jitter
        self.scale_factors = scale_factors
        self.crop_size = crop_size

        super(VOCTrainAuger, self).__init__(dataset, verbosity, pool_size, slim, rand_seed_log)

    def build_graph(self):
        super(VOCTrainAuger, self).build_graph()

        # * Color jittering
        rand_color_jitter = RandColorJitter() if self.is_color_jitter else RandColorJitter(jitter_prob=0.0)
        self.graph.add_node(rand_color_jitter, inputs=['img'], outputs=['jitter_img'])

        # * Rand scale
        rand_scale = RandScale(self.scale_factors)
        self.graph.add_node(rand_scale, inputs=['jitter_img', 'label'], outputs=['scaled_img', 'scaled_label'])

        self.graph.add_node(int_img2float32_img, inputs=['scaled_img'], outputs=['float_img'])

        self.graph.add_node(centralize, inputs=['float_img', {'mean': self.dataset.mean_bgr}],
                            outputs=['centralized_img'])

        self.graph.add_node(pad_img_label,
                            inputs=['centralized_img', 'scaled_label', {'pad_img_to': self.crop_size}],
                            kwargs={'pad_location': 'center', 'ignore_label': self.dataset.ignore_label},
                            outputs=['padded_img', 'padded_label'])

        # * Rand crop
        rand_crop = RandCrop(self.crop_size)
        self.graph.add_node(rand_crop, inputs=['padded_img', 'padded_label'], outputs=['cropped_img', 'cropped_label'])

        # * Mirror
        mirror = RandMirror() if not self.is_multi_mirror else MultiMirror()
        self.graph.add_node(mirror, inputs=['cropped_img', 'cropped_label'], outputs=['mirrored_img', 'mirrored_label'])

        self.graph.add_node(HWC2CHW, inputs=['mirrored_img'], outputs=['CHW_img'])

        self.graph.add_node(collect_example,
                            inputs=['img_id', 'CHW_img', 'mirrored_label'], outputs=['VOCTrainAuger_output'])


class VOCClsTrainAuger(VOCTrainAuger):
    """VOCTrain Auger with class in label

    The output of auger will be (img_id, img, label, cls_in_label). The cls_in_label is an binary vector with length
    class_num, where 1 means class exits in label, else does not exit.
    """
    def build_graph(self):
        super(VOCClsTrainAuger, self).build_graph()

        self.graph.add_node(attach_cls, inputs=['VOCTrainAuger_output'], outputs=['VOCTrainAugerCls_output'])


class VOCTestAuger(_VOCBaseAuger):
    """DataAuger for VOC Dataset when training"""

    def __init__(self, dataset: Union[VOC, VOCAug],
                 verbosity: int=0, pool_size: int=0):
        """DataAuger for VOC Dataset when Test. The img will be centralized and transposed to (C, H, W)

        Args:
            dataset: VOC or VOCAug dataset
            verbosity: See alchemy_cat.data.DataAuger
            pool_size: See alchemy_cat.data.DataAuger
        """
        super(VOCTestAuger, self).__init__(dataset, verbosity, pool_size, slim=True, rand_seed_log=None)

    def build_graph(self):
        super(VOCTestAuger, self).build_graph()

        self.graph.add_node(int_img2float32_img, inputs=['img'], outputs=['float_img'])

        self.graph.add_node(centralize, inputs=['float_img', {'mean': self.dataset.mean_bgr}],
                            outputs=['centralized_img'])

        self.graph.add_node(HWC2CHW, inputs=['centralized_img'], outputs=['CHW_img'])

        self.graph.add_node(collect_example, inputs=['img_id', 'CHW_img', 'label'], outputs=['VOCTestAuger_output'])


class VOCClsTestAuger(VOCTestAuger):
    """VOCTest Auger with class in label

    The output of auger will be (img_id, img, label, cls_in_label). The cls_in_label is an binary vector with length
    class_num, where 1 means class exits in label, else does not exit.
    """
    def build_graph(self):
        super(VOCClsTestAuger, self).build_graph()

        self.graph.add_node(attach_cls, inputs=['VOCTestAuger_output'], outputs=['VOCTestAugerCls_output'])


if __name__ == '__main__':
    voc_aug = VOCAug(split='val')
    voc_aug_auger = VOCClsTrainAuger(voc_aug,
                                     slim=True,
                                     is_multi_mirror=False, is_color_jitter=True, scale_factors=[0.5, 1.0, 1.5])

    mirror_node = None
    for node in voc_aug_auger.rand_nodes:
        if isinstance(node._fct, RandMirror):
            mirror_node = node

    mirror_count = [0] * 2
    scale_factor_count = [0] * 3
    for index, (origin_example, augered_example) in enumerate(zip(voc_aug, voc_aug_auger)):
        origin_img_id, origin_img, origin_label = origin_example
        augered_img_id, augered_img, augered_label, cls_in_label = augered_example

        assert origin_img_id == augered_img_id

        # * Test rand mirror
        if mirror_node._fct.rand_seed == 1:
            mirror_count[0] += 1
        elif mirror_node._fct.rand_seed == -1:
            mirror_count[1] += 1
        else:
            raise ValueError()

        # * Test rand scale
        if int(origin_img.shape[0] * 0.5) == voc_aug_auger.graph.data['scaled_img'].shape[0]:
            scale_factor_count[0] += 1
        elif int(origin_img.shape[0] * 1.0) == voc_aug_auger.graph.data['scaled_img'].shape[0]:
            scale_factor_count[1] += 1
        elif int(origin_img.shape[0] * 1.5) == voc_aug_auger.graph.data['scaled_img'].shape[0]:
            scale_factor_count[2] += 1
        else:
            raise ValueError()

        # * Test Crop Size
        assert augered_img.shape[1:] == (321, ) * 2
        assert augered_img.shape[1:] == augered_label.shape

        # * Test cls_in_label
        assert len(cls_in_label) == 21

        # * Recover img
        if index % 100 == 0:
            recovered_img = (CHW2HWC(augered_img) + np.array(VOC.mean_bgr)).astype(np.uint8)

            img_wall = RowFigureWall([origin_img, recovered_img],
                                     space_width=20, pad_location='center', color_channel_order='BGR')
            label_wall = RowFigureWall(map(label_map2color_map, [origin_label, augered_label]),
                                       space_width=20, pad_location='center', color_channel_order='RGB')

            show_wall = ColumnFigureWall([img_wall, label_wall], space_width=20)
            show_wall.plot()
            plt.show()

            print([VOC_CLASSES[cls_idx] for cls_idx in np.nonzero(cls_in_label)[0]])

    voc_aug_auger = VOCClsTestAuger(voc_aug)

    for index, (origin_example, augered_example) in enumerate(zip(voc_aug, voc_aug_auger)):
        origin_img_id, origin_img, origin_label = origin_example
        augered_img_id, augered_img, augered_label, cls_in_label = augered_example

        assert origin_img_id == augered_img_id

        # * Test cls_in_label
        assert len(cls_in_label) == 21

        # * Recover img
        recovered_img = (CHW2HWC(augered_img) + np.array(VOC.mean_bgr)).astype(np.uint8)

        if index % 100 == 0:

            img_wall = RowFigureWall([origin_img, recovered_img],
                                     space_width=20, pad_location='center', color_channel_order='BGR')
            label_wall = RowFigureWall(map(label_map2color_map, [origin_label, augered_label]),
                                       space_width=20, pad_location='center', color_channel_order='RGB')

            show_wall = ColumnFigureWall([img_wall, label_wall], space_width=20)
            show_wall.plot()
            plt.show()

            print([VOC_CLASSES[cls_idx] for cls_idx in np.nonzero(cls_in_label)[0]])

        assert np.all(recovered_img == origin_img)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/11/25 20:51
@File    : funcs.py
@Software: PyCharm
@Desc    : 
"""
import numpy as np

__all__ = ["func_a", "func_b", "func_c", "func_d"]


def func_a(goo):
    return goo


def func_b(goo):
    return func_a(goo)


def func_c(goo):
    np.array(goo)


def func_d():
    return func_d



# -- configs/mnist/base/cfg.py --

import torchvision.models as model
import torchvision.transforms as T
from torch import optim
from torchvision.datasets import MNIST

from alchemy_cat.dl_config import Config, DEP

cfg = Config()

cfg.rand_seed = 0

# -* Set datasets.
cfg.dt.cls = MNIST
cfg.dt.ini.root = '/tmp/data'
cfg.dt.ini.transform = T.Compose([T.Grayscale(3), T.ToTensor(), T.Normalize((0.1307,), (0.3081,)),])

# -* Set data loader.
cfg.loader.ini.batch_size = 128
cfg.loader.ini.num_workers = 2

# -* Set model.
cfg.model.cls = model.resnet18
cfg.model.ini.num_classes = DEP(lambda c: len(c.dt.cls.classes))

# -* Set optimizer.
cfg.opt.cls = optim.AdamW
cfg.opt.ini.lr = DEP(lambda c: c.loader.ini.batch_size // 128 * 0.01)  # Linear scaling rule.

# -* Set scheduler.
cfg.sched = Config('configs/addons/linear_warm_cos_sched.py')

# -* Set logger.
cfg.log.save_interval = DEP(lambda c: c.sched.epochs // 5, priority=1)  # Save model at every 20% of total epochs.



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: timer.py
@time: 2020/1/11 1:40
@desc:
"""
import time
from timeit import default_timer

from typing import Optional

__all__ = ['Timer']


class Timer(object):
    __unitfactor = {'s': 1,
                    'ms': 1000,
                    'us': 1000000}

    def __init__(self, unit: Optional[str]=None, precision: int=4, name: str="Default Timer"):
        """Timer to record the time interval in program

        Args:
            unit: Timer's unit. Can be s, ms, us. If not given, then the program will adaptively choose a unit.
            precision: Timer's output precision. The total time's decimal part will be rounded according to
                this precision.
            name: Timer's name.
        """
        if unit is not None and unit not in Timer.__unitfactor:
            raise ValueError('Unsupported time unit.')
        self._start = 0
        self._end = 0
        self._total = None
        self._unit = unit
        self._precision = precision
        self.name = name
    
    @property
    def total(self):
        assert self._total is not None
        return self._total

    def start(self):
        self._start = default_timer()
        return self

    def __enter__(self):
        return self.start()

    def close(self):
        self._end = default_timer()
        self._total = (self._end - self._start)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        if self._total is None:
            print(f"Timer {self.name} survived for {self}")

    def __repr__(self):
        if self._total is None:
            total = default_timer() - self._start
        else:
            total = self._total

        if self._unit is None:
            if total > 300.:
                total_s = int(total)
                ret = f"{total_s // 3600}h:{(total_s % 3600) // 60:02}m:{total_s % 60:02}s" + \
                      f"{round(total - total_s, self._precision)}"[1:]
            else:
                global factor, unit
                for unit, factor in Timer.__unitfactor.items():
                    if total * factor > 1.:
                        break
                ret = f"{round(total * factor, self._precision)}{unit}"
        else:
            total *= Timer.__unitfactor[self._unit]
            total = round(total, self._precision)
            ret = f"{total}{self._unit}"

        return f"{self.name}: " + ret


if __name__ == "__main__":
    timer = Timer().start()
    timer.close()
    print(timer)

    with Timer(precision=5) as timer:
        tmp = 1.
        for i in range(10000000):
            if i % 1000000 == 0:
                print(timer)
            tmp *= float(i)
    print(timer)

    timer = Timer(unit='us').start()
    tmp = 1.
    for i in range(1000000):
        tmp *= float(i)
    timer.close()
    print(timer.name, timer)

    timer = Timer(name="deleted Timer")
    time.sleep(2)
    del timer

    def foo():
        timer = Timer(name="unclosed Timer")
        time.sleep(2)

    foo()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: average_value_tracker.py
@time: 2020/2/28 12:43
@desc: Modified from torchnet.meter.averagevaluemeter and torchnet.meter.movingaveragevaluemeter
"""
import math
import numpy as np
import torch

from alchemy_cat.py_tools import Tracker, Statistic

__all__ = ['AverageValueTracker', 'MovingAverageValueTracker']


class AverageValueTracker(Tracker):
    def __init__(self):
        super(AverageValueTracker, self).__init__()
        self._n = None
        self._sum = None
        self._var = None
        self._val = None
        self._mean =  None
        self._mean_old = None
        self._m_s = None
        self._std = None
        self._last = None

        self.reset()
        self._val = 0

    def _add(self, value, n=1):
        self._last = value

        self._val = value
        self._sum += value
        self._var += value * value
        self._n += n

        if self._n == 0:
            self._mean, self._std = np.nan, np.nan
        elif self._n == 1:
            self._mean = 0.0 + self._sum  # This is to force a copy in torch/numpy
            self._std = np.inf
            self._mean_old = self._mean
            self._m_s = 0.0
        else:
            self._mean = self._mean_old + (value - n * self._mean_old) / float(self._n)
            self._m_s += (value - self._mean_old) * (value - self._mean)
            self._mean_old = self._mean
            self._std = np.sqrt(self._m_s / (self._n - 1.0))

    def value(self):
        return self._mean, self._std

    def reset(self):
        self._n = 0
        self._sum = 0.0
        self._var = 0.0
        self._val = 0.0
        self._mean = np.nan
        self._mean_old = 0.0
        self._m_s = 0.0
        self._std = np.nan
        self._last = None

    @property
    def last(self):
        return self._last

    def update(self, value, n=1):
        super(AverageValueTracker, self).update(value, n)
        self._add(value, n)

    @Statistic
    def mean(self):
        return self.value()[0]

    @Statistic
    def std(self):
        return self.value()[1]


class MovingAverageValueTracker(Tracker):
    def __init__(self, window_size):
        super(MovingAverageValueTracker, self).__init__()
        self.window_size = window_size

        self._var = None
        self._n = None
        self._sum = None
        self._last = None
        self._value_queue = torch.zeros(window_size)

        self.reset()

    def reset(self):
        self._sum = 0.0
        self._n = 0
        self._var = 0.0
        self._last = None
        self._value_queue.fill_(0)

    def _add(self, value):
        self._last = value

        queue_id = (self._n % self.window_size)
        old_value = self._value_queue[queue_id]
        self._sum += value - old_value
        self._var += value * value - old_value * old_value
        self._value_queue[queue_id] = value
        self._n += 1

    @property
    def last(self):
        return self._last

    def value(self):
        n = min(self._n, self.window_size)
        mean = self._sum / max(1, n)
        std = math.sqrt(max((self._var - n * mean * mean) / max(1, n - 1), 0))
        return mean, std

    def update(self, value):
        super(MovingAverageValueTracker, self).update(value)
        self._add(value)

    @Statistic
    def mean(self):
        return self.value()[0]

    @Statistic
    def std(self):
        return self.value()[1]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: utils.py
@time: 2020/1/26 23:08
@desc:
"""
from collections.abc import Iterator

from typing import Any, Tuple, Iterable, Callable, Union

from alchemy_cat.py_tools.type import is_int, is_float, totuple


__all__ = ['accumulate', 'size2HW', 'color2scalar']


# Taken from python 3.5 docs
def accumulate(iterable: Iterable, fn: Callable=lambda x, y: x + y):
    'Return running totals'
    # _accumulate([1,2,3,4,5]) --> 1 3 6 10 15
    # _accumulate([1,2,3,4,5], operator.mul) --> 1 2 6 24 120
    it = iter(iterable)
    try:
        total = next(it)
    except StopIteration:
        return
    yield total
    for element in it:
        total = fn(total, element)
        yield total


def size2HW(size: Any) -> Tuple[int, int]:
    """Convert size to H, W

    Args:
        size (Any): If size is int, return (size, size), else return (size[0], size[1])

    Returns: (H, W) converted from size
    """
    if is_int(size):
        if size < 0:
            raise ValueError(f"size{size} must >= 0")
        return int(size), int(size)
    if isinstance(size, Iterator):
        return size2HW(list(size))
    elif len(size) == 2 and is_int(size[0]) and is_int(size[1]):
        if size[0] < 0 or size[1] < 0:
            raise ValueError(f"H, W of size {size} must >= 0")
        return int(size[0]), int(size[1])
    else:
        raise ValueError("size should be int or [int, int] or (int, int)")


def color2scalar(val: Union[int, float, Iterable]) -> Tuple:
    """Convert color value to color scalar

    Args:
        val (Union[int, float, Iterable]): If val is int or float, return (val, val, val), if value is Iterable with 3
            element, return totuple(val), else raise error

    Returns: tuple color scalar
    """
    if is_int(val) or is_float(val):
        img_pad_scalar = (val,) * 3
    else:
        img_pad_scalar = totuple(val)
        if len(img_pad_scalar) != 3:
            raise ValueError(f"If img_pad_value {val} is scalar, it should be a 3-channel color scalar")

    return img_pad_scalar


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/5/6 22:09
@File    : cfg.py.py
@Software: PyCharm
@Desc    : 
"""
from alchemy_cat.py_tools import Cfg2Tune, Param2Tune, ItemLazy

cfg = config = Cfg2Tune()

cfg.rslt_dir = 'itemlazy_config'


def b_get1(c):
    return c.foo0.foo1.a * 2


def b_get2(c):
    return int(c.c)


def lazy_get(c):
    return c.foo0.a_max['b'] * c.foo0.foo1.a


# * Param on same Cfg2Tune
cfg.foo0.foo1.a = Param2Tune([0, 1, 2])
cfg.foo0.foo1.b = Param2Tune([ItemLazy(b_get1, rel=False), ItemLazy(b_get2, rel=False)])
# * Param on diff Cfg2Tune
cfg.foo0.fix = Param2Tune(['0'])
# * Param on root Cfg2Tune
cfg.c = Param2Tune([False, True])
# * Static Value
cfg.foo0.a_max = {'a': 1, 'b': 2}
# * Item Lazy
cfg.foo1.lazy = ItemLazy(lazy_get)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/11/8 21:16
@File    : decorators.py
@Software: PyCharm
@Desc    : 
"""
from warnings import warn

from .timer import Timer
from .color_print import *

__all__ = ['deprecated', 'func_runtime_timer']


def deprecated(func):
    """Show deprecation warning of a function."""

    def wrapper(*args, **kwargs):
        warn('Func {} is now deprecated!'.format(func.__name__))
        return wrapper(*args, **kwargs)

    return wrapper


def func_runtime_timer(offset: int = 40, show_interval: int = 10):
    def decorator(func):
        cons = []

        def func_wrapper(*args, **kwargs):
            space = offset * " "
            nonlocal cons
            with Timer(unit='ms', precision=2, name=f'{func.__name__}耗时') as t:
                result = func(*args, **kwargs)
            bprint(f"{space}| {t}")
            cons.append(t.total)
            if cons and len(cons) % show_interval == 0:
                bprint(f"{space}| -------------------------------------------------------\n"
                       f"{space}| {func.__name__}已经执行{len(cons)}次。\n"
                       f"{space}| 平均耗时{round(sum(cons) / len(cons) * 1000, 2)}ms，\n"
                       f"{space}| 最大耗时{round(max(cons) * 1000, 2)}ms，\n"
                       f"{space}| 最小耗时{round(min(cons) * 1000, 2)}ms。\n"
                       f"{space}| -------------------------------------------------------")
            return result

        return func_wrapper

    return decorator



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/3/5 14:25
@File    : gen_cls_lb.py
@Software: PyCharm
@Desc    : 
"""
import pickle

import numpy as np
from tqdm import tqdm

from alchemy_cat.contrib.voc import VOCAug, VOC_CLASSES


def lb2cls_lb(label: np.ndarray) -> np.ndarray:
    return (np.bincount(label.ravel(), minlength=VOCAug.ignore_label + 1) != 0).astype(np.uint8)[:len(VOC_CLASSES)]


# * 获取来自分割标签的图像标签。
seg_cls_labels = {}
no_bg_image_ids = []

for img_id, _, lb in tqdm(VOCAug('datasets', split='trainval_aug'), dynamic_ncols=True):
    seg_cls_labels[img_id] = lb2cls_lb(lb)

    if seg_cls_labels[img_id][0] == 0:
        no_bg_image_ids.append(img_id)

with open('datasets/VOC2012/third_party/seg_cls_labels.pkl', 'wb') as pkl_f:
    pickle.dump(seg_cls_labels, pkl_f)

with open('datasets/VOC2012/third_party/no_bg_image_ids.pkl', 'wb') as pkl_f:
    pickle.dump(no_bg_image_ids, pkl_f)

# * 获取来自检测标签的图像标签。
ref_det_cls_labels = np.load('datasets/voc_cls_labels.npy', allow_pickle=True).tolist()
det_cls_labels = {}

for int_img_id, ref_det_cls_lb in tqdm(ref_det_cls_labels.items(), dynamic_ncols=True):
    det_cls_labels[f'{(sid := str(int_img_id))[:4]}_{sid[4:]}'] = np.insert(ref_det_cls_lb.astype(np.uint8), 0, 1)

with open('datasets/VOC2012/third_party/det_cls_labels.pkl', 'wb') as pkl_f:
    pickle.dump(det_cls_labels, pkl_f)

# * 获取混合标签。
# seg_cls_lb与det_cls_lb对比发现：
# 1. 大约有818张图片，det_cls_lb中的困难物体没有标出。除此以外，基本相同。
# 2. bg=0的图片有10张，的确是没有背景。
difficult_image_ids = []
ignore_diff_cls_labels = {}
for img_id in det_cls_labels.keys():
    if np.any(diff_mask := seg_cls_labels[img_id][1:] != det_cls_labels[img_id][1:]):  # 背景遵循seg_cls_label。
        difficult_image_ids.append(img_id)

    ignore_diff_cls_lb = seg_cls_labels[img_id].copy()
    ignore_diff_cls_lb[1:][diff_mask] = 255
    ignore_diff_cls_labels[img_id] = ignore_diff_cls_lb

with open('datasets/VOC2012/third_party/difficult_image_ids.pkl', 'wb') as pkl_f:
    pickle.dump(difficult_image_ids, pkl_f)

with open('datasets/VOC2012/third_party/ignore_diff_cls_labels.pkl', 'wb') as pkl_f:
    pickle.dump(ignore_diff_cls_labels, pkl_f)



from alchemy_cat.dag.io import Input, Output
import pytest
from alchemy_cat.dag.errors import PyungoError


def test_Input():
    inp = Input(name='a')
    assert inp.name == 'a'
    assert inp.value is None
    assert inp.is_kwarg is False
    assert inp.is_constant is False


def test_Input_kwarg():
    inp = Input(name='a', is_kwarg=True)
    assert inp.name == 'a'
    assert inp.value is None
    assert inp.is_kwarg is True
    assert inp.is_constant is False


def test_Input_constant():
    inp = Input(name='a', value=2)
    assert inp.name == 'a'
    assert inp.value == 2
    assert inp.is_kwarg is False
    assert inp.is_constant is True


def test_Input_constant_kwarg():
    inp = Input(name='a', value=2, is_kwarg=True)
    assert inp.name == 'a'
    assert inp.value == 2
    assert inp.is_kwarg is True
    assert inp.is_constant is True


def test_Output():
    out = Output('a')
    assert out.name == 'a'
    out.value = 2
    assert out._value == 2
    assert out.value == 2


def test_io_maps():
    for Class in [Input, Output]:
        io = Class(name='a')
        assert io.map == 'a'
        io = Class(name='a', map='b')
        assert io.map == 'b'

def test_dot_in_name():
    with pytest.raises(PyungoError) as err:
        inp = Input(name='a.b')

    assert "IO can't have '.' in name, however get name = a.b" in str(err.value)

def test_dot_in_map():
    with pytest.raises(PyungoError) as err:
        inp = Input(name='a', map='a.b')

    assert "IO can't have '.' in map, however get map = a.b" in str(err.value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2024/7/2 14:53
@File    : utils.py
@Software: PyCharm
@Desc    : 
"""
import typing as t

import torch
import torch.nn.functional as F
import torchvision.transforms as T
from torch import nn
from torch.utils.data import DataLoader
from torchvision.datasets import MNIST

from alchemy_cat.dl_config import ADict

__all__ = ['eval_model']


class RetDict(t.TypedDict):
    test_loss: float
    acc: float


def eval_model(model: nn.Module) -> RetDict:
    test_loader = DataLoader(MNIST('/tmp/data',
                                   train=False,
                                   transform=T.Compose([T.Grayscale(3),
                                                        T.ToTensor(),
                                                        T.Normalize((0.1307,), (0.3081,)),])),
                             batch_size=512, shuffle=False, num_workers=2, drop_last=False)
    N = len(t.cast(t.Sized, test_loader.dataset))

    model.eval()

    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to('cuda'), target.to('cuda')
            output = model(data)
            test_loss += F.cross_entropy(output, target, reduction='sum').item()  # sum up batch loss
            pred = output.argmax(dim=1, keepdim=True)  # get the index of the max log-probability
            correct += pred.eq(target.view_as(pred)).sum().item()

    test_loss /= N
    acc = 100. * correct / N

    print(f'\nTest set: Average loss: {test_loss:.4f}, Accuracy: {correct}/{N} ({acc:.0f}%)\n')

    model.train()

    ret: RetDict = ADict()
    ret.test_loss = test_loss
    ret.acc = acc

    return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/5/20 22:57
@File    : utils.py
@Software: PyCharm
@Desc    : 
"""
from typing import Tuple, Callable

import numpy as np
import torch

from alchemy_cat.alg import channel_min_max_norm

__all__ = ["gather_segment", "scatter_segment", "cam_clip_norm_score", "get_fg_cam"]


def gather_segment(arr: np.ndarray, segment: np.ndarray,
                   gather_func: Callable[[np.ndarray], np.ndarray] = lambda vec: np.mean(vec, axis=1)) \
        -> Tuple[np.ndarray, np.ndarray]:
    """Gather array's vector in each segment with gather func.

    Args:
        arr: (C, H, W) arr to be gathered.
        segment: (H, W) segment mask, each segment has it's unique int label.
        gather_func: Callable to gather (C, segment_area) arr in segment to a vector with length C.

    Returns:
        (C, segment_num) gathered segment vector, (segment_num, ) segment areas.
    """
    segment_indices = np.unique(segment)

    segment_vecs = []
    segment_areas = []

    for segment_idx in segment_indices:
        vec_in_segment = arr[:, segment == segment_idx]
        segment_areas.append(vec_in_segment.shape[1])
        segment_vecs.append(gather_func(vec_in_segment))

    return np.stack(segment_vecs, axis=1), np.array(segment_areas, dtype=np.int64)


def scatter_segment(segment_vec: np.ndarray, segment: np.ndarray) -> np.ndarray:
    """Scatter segment_vec to each segment.

    Args:
        segment_vec: (C, segment_num) segment vector.
        segment: (H, W) segment mask, each segment has it's unique int label.

    Returns:
        (C, H, W) arr scattered from segment_vec according to segment.
    """
    scattered_arr = np.zeros((segment_vec.shape[0], ) + segment.shape, dtype=segment_vec.dtype)

    segment_indices = np.unique(segment)

    for i, segment_idx in enumerate(segment_indices):
        scattered_arr[:, segment == segment_idx] = segment_vec[:, i][:, None]

    return scattered_arr


def cam_clip_norm_score(cam: np.ndarray) -> np.ndarray:
    clip_cam = np.clip(cam, .0, None)
    return channel_min_max_norm(torch.from_numpy(clip_cam[None, ...]), 1e-5).numpy()[0]


def get_fg_cam(cam: np.ndarray, cls_in_label: np.ndarray):
    fore_in_label = cls_in_label[1:]
    return cam[fore_in_label.astype(bool)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: test_tracker.py
@time: 2020/3/13 14:48
@desc:
"""
import pytest
import numpy as np
import copy
from sklearn.metrics import f1_score
from tqdm import tqdm
from colorama import Style, Fore

from alchemy_cat.contrib.voc import VOCAug, VOC_CLASSES
from alchemy_cat.data.plugins.augers import scale_img_label
from alchemy_cat.contrib.metrics import SegmentationMetric
from alchemy_cat.py_tools.tracker import Tracker, Statistic
from alchemy_cat.py_tools.type import dict_with_arr_is_eq

kScaleFactor = 0.125


def _get_conf_matrix(label_true, label_pred, n_class):
    mask = (label_true >= 0) & (label_true < n_class)
    conf_matrix = np.bincount(
        n_class * label_true[mask].astype(int) + label_pred[mask],
        minlength=n_class ** 2,
    ).reshape(n_class, n_class)
    return conf_matrix


def scores(label_trues, label_preds, n_class):
    """Copied and modified from https://github.com/kazuto1011/deeplab-pytorch as Reference Metrics.
        Supposed to be correct"""
    conf_matrix = np.zeros((n_class, n_class))
    for lt, lp in zip(label_trues, label_preds):
        conf_matrix += _get_conf_matrix(lt.flatten(), lp.flatten(), n_class)

    condition_positive = np.sum(conf_matrix, axis=1)
    acc = np.diag(conf_matrix).sum() / conf_matrix.sum()
    cls_recall = np.diag(conf_matrix) / conf_matrix.sum(axis=1)
    macro_recall = np.nanmean(cls_recall)
    cls_precision = np.diag(conf_matrix) / np.sum(conf_matrix, axis=0)
    macro_precision = np.nanmean(cls_precision)
    cls_F_score = (2 * cls_precision * cls_recall) / (cls_precision + cls_recall)
    F_score = np.nanmean(cls_F_score)

    cls_IoU = np.diag(conf_matrix) / (conf_matrix.sum(axis=1) + conf_matrix.sum(axis=0) - np.diag(conf_matrix))
    valid = conf_matrix.sum(axis=1) > 0  # added, if some class doesn't exit, then don't take it into count
    mean_IoU = np.nanmean(cls_IoU[valid])
    freq = conf_matrix.sum(axis=1) / conf_matrix.sum()
    freq_weighted_IoU = (freq[freq > 0] * cls_IoU[freq > 0]).sum()

    return {
        "Conf Matrix": conf_matrix,
        "Frequency Weighted IoU": freq_weighted_IoU,
        "Mean IoU": mean_IoU,
        "Class IoU": cls_IoU,
        "Accuracy": acc,
        "Macro Recall": macro_recall,
        "Class Recall": cls_recall,
        "Class Precision": cls_precision,
        "Macro Precision": macro_precision,
        "F Score": F_score,
        "Class Samples Num": condition_positive
    }


@pytest.fixture(scope='module')
def voc_dataset():
    return VOCAug(split='val')


@pytest.fixture(scope='module')
def metric_results():
    # * Tracker
    seg_metric = SegmentationMetric(class_names=VOC_CLASSES)

    # * For scores
    preds, gts = [], []

    for example in tqdm(VOCAug(split='val'), desc='Metric Progress', unit='samples', dynamic_ncols=True):
        _, img, label = example
        label = copy.deepcopy(label)
        img_shape = img.shape[:2]

        # * Scale-Recover label as pred
        scaled_img, scaled_label = scale_img_label(kScaleFactor, img, label)
        _, recovered_label = \
            scale_img_label(1, scaled_img, scaled_label, (lambda _: img_shape[0], lambda _: img_shape[1]))
        # label[recovered_label == 255] = 255
        recovered_label[(label != 255) & (recovered_label == 255)] = 0 # preds don't have ignore label

        # * Update tracker
        seg_metric.update(recovered_label, label)

        # * Update for score
        preds.append(recovered_label); gts.append(label)

    # * Calculate reference metrics
    ref_metrics = scores(gts, preds, len(VOC_CLASSES))

    return seg_metric, ref_metrics, preds, gts


@pytest.mark.tryfirst
def test_statistic_cache(metric_results):
    seg_metric, _, _, _ = metric_results
    seg_metric = copy.deepcopy(seg_metric)

    calculated_statistic = seg_metric.statistics()

    tracker_classes = [base_cls for base_cls in seg_metric.__class__.mro() if issubclass(base_cls, Tracker)]

    statistics = [statistic
            for cls in tracker_classes
            for statistic in cls.__dict__.values() if isinstance(statistic, Statistic)]

    for statistic in statistics:
        statistic.fget = None

    cached_statistic = seg_metric.statistics()

    # * Same length
    assert len(calculated_statistic) == len(cached_statistic)

    # * Save keys and values
    dict_with_arr_is_eq(calculated_statistic, cached_statistic)

    for cal_item, cached_item in zip(calculated_statistic.items(), cached_statistic.items()):
        cal_key, cal_value = cal_item
        cached_key, cached_val = cached_item

        assert cal_key == cached_key

        if isinstance(cal_value, np.ndarray):
            assert np.all(cal_value == cached_val)
        else:
            assert cal_value == cached_val


def test_statistics_calculation(metric_results):
    seg_metric, ref_metrics, preds, gts = metric_results

    assert np.all(seg_metric.conf_matrix == ref_metrics['Conf Matrix'])

    assert seg_metric.weighted_mIoU == ref_metrics['Frequency Weighted IoU']

    assert seg_metric.mIoU == ref_metrics['Mean IoU']

    assert np.all(seg_metric.cls_IoU == ref_metrics['Class IoU'])

    assert seg_metric.ACC == ref_metrics['Accuracy']

    assert seg_metric.macro_avg_recall == ref_metrics['Macro Recall']

    assert np.all(seg_metric.cls_recall == ref_metrics['Class Recall'])

    assert np.all(seg_metric.cls_precision == ref_metrics['Class Precision'])

    assert seg_metric.macro_avg_precision == ref_metrics['Macro Precision']

    assert seg_metric.F1_score == ref_metrics['F Score']

    assert np.all(seg_metric.condition_positive == ref_metrics['Class Samples Num'])

    preds, gts = np.concatenate(preds, axis=None), np.concatenate(gts, axis=None)
    mask = gts != 255
    preds, gts = preds[mask], gts[mask]

    # ORZ: It costs 1.5min to calculate f1_score. That's why we don't use ski-learn to calculate metric in alchemy_cat
    assert seg_metric.F1_score == f1_score(gts, preds, average='macro')

    assert seg_metric.weighted_F1_score == pytest.approx(f1_score(gts, preds, average='weighted'), rel=1e-6)


def test_update_count(metric_results, voc_dataset):
    seg_metric, _, _, _ = metric_results

    assert seg_metric.update_count == len(voc_dataset)


def test_reset(metric_results):
    seg_metric, _, _, _ = metric_results

    origin_init_dict = seg_metric.init_dict

    seg_metric.reset()

    assert seg_metric.update_count == 0

    assert len(seg_metric._statistic_status) == 0

    assert seg_metric.class_num == len(VOC_CLASSES)

    assert seg_metric.class_names == VOC_CLASSES

    assert np.all(seg_metric.conf_matrix == np.zeros((len(VOC_CLASSES), len(VOC_CLASSES)), dtype=np.int32))

    assert np.all(seg_metric._last_conf_matrix == np.zeros((len(VOC_CLASSES), len(VOC_CLASSES)), dtype=np.int32))

    assert dict_with_arr_is_eq(origin_init_dict, seg_metric.init_dict)


def test_show(metric_results):
    seg_metric, _, _, _ = metric_results
    seg_metric.show_conf_matrix()

    print(Fore.LIGHTRED_EX + "\n Need Visual judgment!" + Style.RESET_ALL)


def test_importance_filter(metric_results):
    seg_metric, _, _, _ = metric_results

    level0 = seg_metric.statistics(importance=0)
    for name, _ in level0.items():
        assert getattr(SegmentationMetric, name).importance >= 0

    level1 = seg_metric.statistics(importance=1)
    for name, _ in level1.items():
        assert getattr(SegmentationMetric, name).importance >= 1

    level2 = seg_metric.statistics(importance=2)
    for name, _ in level2.items():
        assert getattr(SegmentationMetric, name).importance >= 2


def test_statistic_order(metric_results):
    seg_metric, _, _, _ = metric_results

    keys = list(seg_metric.statistics().keys())

    for i in range(len(keys) - 1):
        for j in range(i + 1, len(keys)):
            assert getattr(SegmentationMetric, keys[i]).importance <= getattr(SegmentationMetric, keys[j]).importance
            if getattr(SegmentationMetric, keys[j]).importance == getattr(SegmentationMetric, keys[i]).importance:
                assert keys[i] < keys[j]


def test_save_metrics(metric_results):
    seg_metric, _, _, _ = metric_results

    seg_metric.save_metric('./Temp/imp0', importance=0, dpi=400)

    seg_metric.save_metric('./Temp/imp1', importance=1, dpi=400)

    seg_metric.save_metric('./Temp/imp2', importance=2, dpi=400)

    print(Fore.LIGHTRED_EX + "\n Need Visual judgment!" + Style.RESET_ALL)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/4/26 21:58
@File    : check_img_exits.py
@Software: PyCharm
@Desc    : 
"""
from imghdr import what
import os.path as osp

__all__ = ["check_img_exits"]


def check_img_exits(img_path: str) -> bool:
    """检查图像文件是否存在且完整。

    Args:
        img_path: 图像路径。

    Returns:
        返回图像是否存在。
    """
    if osp.isfile(img_path) and what(img_path) is not None:
        return True
    else:
        return False



import numpy as np

from alchemy_cat.dag.core import Graph


def f_my_function(a1, a2):
        return a1 + a2 + 1


class f_my_function2(object):
    def __init__(self, add_or_sub = 'add', factor=1):
        if add_or_sub == 'add':
            self.func = lambda x, y: x + y
        else:
            self.func = lambda x, y: x - y
        self.factor = factor
    def generate_constant(self):
        return 1
    def __call__(self, a1, a2):
        return (self.func(a1, a2) + self.generate_constant()) * self.factor


if __name__ == '__main__':
    graph = Graph()

    def register_graph(graph):
        @graph.register(inputs=['c', 'b'], outputs=['d'], init={'add_or_sub': 'sub'})
        class f_my_function1(object):
            def __init__(self, add_or_sub = 'add', factor=1):
                if add_or_sub == 'add':
                    self.func = lambda x, y: x + y
                else:
                    self.func = lambda x, y: x - y

                self.factor = factor

            def generate_constant(self):
                return 1

            def __call__(self, a1, a2):
                return (self.func(a1, a2) + self.generate_constant()) * self.factor

        graph.add_node(f_my_function2, inputs=['d', 'a'], outputs=['e'], init={'add_or_sub': 'add', 'factor': 2})
        graph.add_node(f_my_function, inputs=['a', 'b'], outputs=['c'])

    register_graph(graph)

    res = graph.calculate(data={'a': 2, 'b': 3})
    assert graph.data['c'] == 6
    assert graph.data['d'] == 4
    assert res == 14
    assert graph.data['e'] == 14

    # make sure it is independent
    res = graph.calculate(data={'a': 2, 'b': 3})
    assert graph.data['c'] == 6
    assert graph.data['d'] == 4
    assert res == 14
    assert graph.data['e'] == 14


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: env.py
@time: 2020/1/16 4:46
@desc:
"""
import os
import os.path as osp
import sys
from pprint import pprint
from typing import Union, Optional, Tuple

import cv2
import torch
import torch.distributed as dist

from ..dl_config import parse_config, Config, ItemLazy
from ..py_tools import get_process_info, get_local_time_str, set_rand_seed, Logger, meow, init_loguru

__all__ = ['get_device', 'init_env']


def get_device(is_cuda: bool=True, cuda_id: int=0, verbosity: bool=True) -> torch.device:
    """Get default device

    Args:
        is_cuda (bool): If false, always use cpu as device
        cuda_id (int): Default cuda device id.
        verbosity (bool): If True, show current devices

    Returns: Default device

    """
    if is_cuda:
        if not torch.cuda.is_available():
            raise ValueError("is_cuda = True while no cuda device is available.")
        if cuda_id >= torch.cuda.device_count():
            raise ValueError(f"cuda_id must < device_count = {torch.cuda.device_count()}")
        if cuda_id < 0:
            raise ValueError(f"cuda_id = {cuda_id} must > 0")

    device = torch.device("cuda:" + str(cuda_id) if is_cuda else "cpu")

    if verbosity:
        if is_cuda:
            print("设备列表:")
            for i in range(torch.cuda.device_count()):
                msg = f"    {i}: {torch.cuda.get_device_name(i)}"
                if i == cuda_id:
                    msg += "，作为默认设备。"
                print(msg)
        else:
            print("设备列表: CPU")
        print("\n")
    return device


def welcome():
    print("\033[32m##################o(*≧▽≦)ツ 圣火昭昭 圣火耀耀 凡我弟子 喵喵喵喵 φ(≧ω≦*)♪##################\033[0m",
          end="\n\n")


def init_env(*, is_cuda: Union[bool, int] = True, is_benchmark: bool = False, is_train: bool = True,
             config_path: str | dict = None, config_root: str='./configs',
             experiments_root: str = "experiment", rand_seed: Union[bool, str, int] = False,
             cv2_num_threads: int = -1,
             verbosity: bool = True, log_stdout: Union[bool, str] = False, loguru_ini: bool | dict=True,
             local_rank: Optional[int] = None, silence_non_master_rank: Optional[bool] = False,
             reproducibility: Optional[bool] = False,
             is_debug: bool=False) \
        -> Tuple[torch.device, Optional[Config]]:
    """Init torch training environment

    Args:
        is_cuda (Optional(bool, int)): If False, always use CPU. If Ture, use GPU and set GPU:0 as default device. If
            int, set GPU:i as default device.
        is_benchmark (bool): If True, set torch.backends.cudnn.benchmark = True
        is_train (bool): If False, disable grad
        config_path (str | Config): The path to config
        config_root: Root dictionary of configs, for auto_rslt_dir only.
        experiments_root (str): The path where experiments result are stored
        rand_seed (Union[bool, str, int]) : If True, fix random of torch, numpy, python's random module from
            config.RAND_SEED. If False, don't fix random. If rand_seed is int or str, fix random according to
            rand_seed. (Default: False)
        cv2_num_threads (int): Set cv2 threads num by cv2.setNumThreads(cv2_num_threads). If < 0,
            don't set. (Default: -1)
        verbosity (bool): If True, print detailed info
        log_stdout (Union[bool, str]): If True, the stdout will be logged to corresponding experiment dir. If False, the
            stdout will not be logged. If log_stdout is str, it will be recognized as a path and stdout will be logged
            to that path. (Default: False)
        loguru_ini (dict): init loguru with loguru_ini. (Default: True)
        local_rank (Optional[int]): If not None, init distributed parallel env with rank = local_rank with
            init_method = "env://".  Default device will also be set as "cuda:local_rank" .Make sure environment
            is pre-set.
        silence_non_master_rank (bool): If True, non-master rank's (rank > 0) print will be silenced. (Default: False)
        reproducibility (bool): If True, do everything possible to make the pytorch program reproducible.
            (Default: False)
        is_debug (bool): If True, meow.is_debug will be set to True. (Default: False)

    Returns: Default device and config (If config_path is None, config is None)
    """
    # * Set distributed, verbosity is delayed.
    if local_rank is not None:
        # ** Check distributed env
        if not dist.is_available():
            raise ValueError(f"local_rank = {local_rank} while torch.distributed is not available")
        if not torch.cuda.is_available():
            raise ValueError(f"init_env only support cuda device distributed with nccl backend. "
                             f"local_rank = {local_rank} while torch.cuda is not available")

        # ** Set cuda device id
        if is_cuda is False:
            raise ValueError(f"When set local rank, cuda is needed. However, is_cuda = {is_cuda}")
        if (isinstance(is_cuda, int) and not isinstance(is_cuda, bool)) and (is_cuda != local_rank):  # bool是int的子类。
            raise ValueError(f"local_rank = {local_rank} must equal to is_cuda = {is_cuda}")

        if isinstance(is_cuda, bool):
            is_cuda = local_rank  # 若is_cuda为bool，则将其设置为local_rank，即默认使用local_rank作为cuda_id。

        # ** set device & init_process_group
        dist.init_process_group(
            backend='nccl',
            init_method="env://",
            rank=local_rank
        )

    # * Read CONFIG, verbosity is delayed.
    config = parse_config(config_path, experiments_root, config_root) if config_path is not None else None
    if meow.cfg is not None:
        raise RuntimeError(f"meow.cfg = {meow.cfg} should be None before init_env")
    meow.cfg = config  # 新瓶装旧酒，旧酒望新瓶。置酒猫头上，配置从我游。

    def get_stdout_log_dir_from_config():
        return osp.join(config.rslt_dir, 'stdout')

    def get_stdout_log_file(stdout_log_dir):
        if local_rank is not None:
            prefix = '.' if dist.get_rank() > 0 else ''
            file_name = prefix + '-'.join(['stdout', f"rank{dist.get_rank()}",
                                           get_local_time_str(for_file_name=True)]) + '.log'
        else:
            file_name = '-'.join(['stdout', get_local_time_str(for_file_name=True)]) + '.log'
        return osp.join(stdout_log_dir, file_name)

    # * Log stdout
    # ** Get log file
    if isinstance(log_stdout, bool):
        stdout_log_file = get_stdout_log_file(get_stdout_log_dir_from_config()) if log_stdout else None
    elif isinstance(log_stdout, str):
        stdout_log_file = get_stdout_log_file(log_stdout)
    else:
        raise ValueError(f"log_stdout: {log_stdout} should be bool or path str")
    # ** Set logger
    if stdout_log_file is not None:
        if local_rank is not None:
            silence = silence_non_master_rank and (dist.get_rank() > 0)
            Logger(stdout_log_file, real_time=False, silence=silence)
            if verbosity:
                print(f"标准输出重定向到：{stdout_log_file}，Silence = {silence}。", end="\n\n")
        else:
            Logger(stdout_log_file, real_time=False)
            if verbosity:
                print(f"标准输出重定向到：{stdout_log_file}。", end="\n\n")

    # -* loguru初始化。
    if loguru_ini:
        init_loguru(**({} if isinstance(loguru_ini, bool) else loguru_ini))

    # * Welcome & Show system info
    if verbosity:
        welcome()
        print(f"当前工作目录为：{os.getcwd()}")
        print(f"当前python搜索路径为：")
        pprint(sys.path)
        print(f"当前进程信息有：")
        pprint(get_process_info(), sort_dicts=False)
        print("\n")

    if verbosity:
        print("\033[32m-------------------------------------- 初始化开始 --------------------------------------\033[0m",
              end="\n\n")

    # * Print distributed delaying verbosity
    if verbosity and (local_rank is not None):
        print(f"已初始化torch分布式环境，默认CUDA设备ID被设置为：local_rank = {local_rank}。 \n"
              f"    Progress Group Rank: {dist.get_rank()}\n"
              f"    World Size: {dist.get_world_size()}\n"
              f"    Local Rank: {local_rank}", end="\n\n")

    # *  Get device
    if isinstance(is_cuda, bool):
        device = get_device(is_cuda, cuda_id=0, verbosity=verbosity)
    elif isinstance(is_cuda, int):
        device = get_device(is_cuda=True, cuda_id=is_cuda, verbosity=verbosity)
    else:
        raise ValueError(f"Parameter is_cuda = {is_cuda} must be str or int")
    if device != torch.device('cpu'):
        torch.cuda.set_device(device)

    # * Set benchmark
    torch.backends.cudnn.benchmark = is_benchmark
    if verbosity:
        print(f"torch.backends.cudnn.benchmark = {is_benchmark}", end="\n\n")

    # * Set cv2 threads num
    if cv2_num_threads >= 0:
        cv2.setNumThreads(cv2_num_threads)
        if verbosity:
            print(f"cv2.setNumThreads({cv2_num_threads})", end="\n\n")

    # * If test, disable grad
    torch.set_grad_enabled(is_train)
    if verbosity:
        print(f"torch.set_grad_enabled({is_train})", end="\n\n")

    def get_rand_seed_from_config():
        if 'rand_seed' not in config:
            raise ValueError(f"CONFIG didn't have key rand_seed")
        return config.rand_seed

    # * Set rand seed
    # ** 若rand_seed不是False，则解析得到rand_seed_ori，反之rand_seed_ori为None。
    if isinstance(rand_seed, bool):
        rand_seed_ori = get_rand_seed_from_config() if rand_seed else None
    elif isinstance(rand_seed, int) or isinstance(rand_seed, str):
        rand_seed_ori = rand_seed
    else:
        raise ValueError(f"rand_seed should be bool, int or str.")

    # ** 若rand_seed_ori不是None（需要设置随机种子点），且分布式训练，则给每个rank设置不同随机种子偏置。反之偏置为None。
    if rand_seed_ori is not None and local_rank is not None:
        # 偏置乘以1000，使不同rank的种子点，即使向后递延（如Dataloader的worker，种子点为s、s+1、···）也不会发生冲突。
        rand_seed_bias = local_rank * 1000 if isinstance(rand_seed_ori, int) else str(local_rank * 1000)
    else:
        rand_seed_bias = None

    # ** 若rand_seed_ori不为None，则设置随机种子点。若rand_seed_bias不为None，则要给种子点加上偏置。
    if rand_seed_ori is not None:
        rand_seed_final = rand_seed_ori if rand_seed_bias is None else rand_seed_ori + rand_seed_bias
        set_rand_seed(rand_seed_final)
        meow.rand_seed_final = rand_seed_final
        if verbosity:
            print(f"设置随机种子为：{rand_seed_final}" +
                  (f" = {rand_seed_ori} + {rand_seed_bias}" if rand_seed_bias is not None else ''), end="\n\n")

    # * Set reproducibility
    if reproducibility:
        # ** 检查是否设置了随机种子点。若没有设置（rand_seed为False，报错）。
        if isinstance(rand_seed, bool) and not rand_seed:
            raise ValueError(f"rand_seed must be set if reproducibility required. ")
        # ** 检查是否设置了benchmark。若设置，则与复现要求冲突，报错。
        if is_benchmark:
            raise ValueError(f"is_benchmark must be False if reproducibility required. ")
        # ** 针对不同版本pytorch，设置其尽量使用确定性算法。
        if hasattr(torch, 'set_deterministic'):
            torch.set_deterministic(True)
        if hasattr(torch, 'use_deterministic_algorithms'):
            torch.use_deterministic_algorithms(True)
        torch.backends.cudnn.deterministic = True
        if verbosity:
            print("设置torch.backends.cudnn.deterministic = True，torch.set_deterministic(True)或"
                  "torch.use_deterministic_algorithms(True)被调用，使算法可复现。", end="\n\n")

    # * Set is_debug
    meow.is_debug = is_debug
    if verbosity:
        print(f"meow.is_debug = {is_debug}", end="\n\n")

    # * End of init
    if verbosity:
        print("\033[32m-------------------------------------- 初始化结束 --------------------------------------\033[0m",
              end="\n\n")

    # * Compute config's ItemLazy.
    if config is not None:
        print("\033[32m-------------------------------------- 计算惰性项 --------------------------------------\033[0m")
        config = ItemLazy.compute_item_lazy(config)
        config.freeze()
        print("\033[32m-------------------------------------- 惰性项算毕 --------------------------------------\033[0m",
              end="\n\n")

    # * Print config's delaying verbosity
    if verbosity and config is not None:
        print("\033[32m-------------------------------------- 配置项开始 --------------------------------------\033[0m")
        print(config.to_txt(prefix="cfg."))
        print("\033[32m-------------------------------------- 配置项结束 --------------------------------------\033[0m",
              end="\n\n")

    # * Return
    return device, config



# -- train.py --
import argparse
import json

import torch
import torch.nn.functional as F
from rich.progress import track
from torch.optim.lr_scheduler import SequentialLR

from alchemy_cat.dl_config import load_config
from utils import eval_model

parser = argparse.ArgumentParser(description='AlchemyCat MNIST Example')
parser.add_argument('-c', '--config', type=str, default='configs/mnist/base,sched_from_addon,2xbs,2÷epo/cfg.py')
args = parser.parse_args()

# Folder 'experiment/mnist/base' will be auto created by `load` and assigned to `cfg.rslt_dir`
cfg = load_config(args.config, experiments_root='/tmp/experiment', config_root='configs')
print(cfg)

torch.manual_seed(cfg.rand_seed)  # Use `cfg` to set random seed

dataset = cfg.dt.cls(**cfg.dt.ini)  # Use `cfg` to set dataset type and its initial parameters

# Use `cfg` to set changeable parameters of loader,
# other fixed parameter like `shuffle` is set in main code
loader = torch.utils.data.DataLoader(dataset, shuffle=True, **cfg.loader.ini)

model = cfg.model.cls(**cfg.model.ini).train().to('cuda')  # Use `cfg` to set model

# Use `cfg` to set optimizer, and get `model.parameters()` in run time
opt = cfg.opt.cls(model.parameters(), **cfg.opt.ini, weight_decay=0.)

# Use `cfg` to set warm and main scheduler, and `SequentialLR` to combine them
warm_sched = cfg.sched.warm.cls(opt, **cfg.sched.warm.ini)
main_sched = cfg.sched.main.cls(opt, **cfg.sched.main.ini)
sched = SequentialLR(opt, [warm_sched, main_sched], [cfg.sched.warm_epochs])

for epoch in range(1, cfg.sched.epochs + 1):  # train `cfg.sched.epochs` epochs
    for data, target in track(loader, description=f"Epoch {epoch}/{cfg.sched.epochs}"):
        F.cross_entropy(model(data.to('cuda')), target.to('cuda')).backward()
        opt.step()
        opt.zero_grad()

    sched.step()

    # If cfg.log is defined, save model to `cfg.rslt_dir` at every `cfg.log.save_interval`
    if cfg.log and epoch % cfg.log.save_interval == 0:
        torch.save(model.state_dict(), f"{cfg.rslt_dir}/model_{epoch}.pth")

    eval_model(model)

if cfg.log:
    eval_ret = eval_model(model)
    with open(f"{cfg.rslt_dir}/eval.json", 'w') as json_f:
        json.dump(eval_ret, json_f)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2024/7/10 18:05
@File    : ac_cfg.py
@Software: PyCharm
@Desc    : 
"""
from alchemy_cat.py_tools import Config

cfg = Config()

cfg.rand_seed = 0
cfg.X_method.method = 'L2'
cfg.lp.ini.k = 50
cfg.lp.ini.cg_guess = False
cfg.some.lst = [1, 2, 3]
cfg.some.lst_dic = [{'a': 1}, {'b': 2}]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: voc2012seg.py
@time: 2020/1/7 17:16
@desc:
"""
import os.path as osp
import pickle
from math import ceil

import cv2
import imagesize
import numpy as np
from PIL import Image
from addict import Dict
from alchemy_cat.acplot.shuffle_ch import BGR2RGB, RGB2BGR
from alchemy_cat.data.dataset import Dataset, Subset

from .utils import label_map2color_map, VOC_CLASSES

__all__ = ['VOC', 'VOCAug', 'VOCAug2']


class _VOCBase(Dataset):
    """
    PASCAL VOC and VOC Aug Segmentation base dataset
    """
    class_names = VOC_CLASSES
    class_num = len(class_names)
    mean_bgr = [104.008, 116.669, 122.675]
    std_bgr = [57.375, 57.12, 58.395]
    ignore_label = 255

    def __init__(self, root: str="./contrib/datasets", year="2012", split: str="train", PIL_read: bool=False,
                 ret_img_file: bool=False):
        """
        Args:
            root (str): The parent dir of VOC dataset
            year (str): "2012"
            split (str): "train"/"val"/"trainval"
            PIL_read (bool): If True, use PIL to read image, else use cv2
            ret_img_file (bool): If True, return image file path
        """
        self.root = root
        self.year = year
        self.split = split
        self.PIL_read = PIL_read
        self.ret_img_file = ret_img_file
        self.files = []
        self.image_ids = []

        self.image_dir: str | None = None
        self.label_dir: str | None = None

        self._set_files()

    def _set_files(self):
        raise NotImplementedError

    def __len__(self):
        return len(self.files)

    def get_by_img_id(self, img_id: str):
        try:
            index = self.image_ids.index(img_id)
        except ValueError:
            raise RuntimeError(f"Can't find img_id {img_id} in dataset's image_ids list")
        return self[index]


class VOC(_VOCBase):
    """
    PASCAL VOC Segmentation dataset
    """
    def _set_files(self):
        self.root = osp.join(self.root, f"VOC{self.year}")
        self.image_dir = osp.join(self.root, "JPEGImages")
        self.label_dir = osp.join(self.root, "SegmentationClass")

        if self.split in ["train", "trainval", "val", "test"]:
            file_list = osp.join(
                self.root, "ImageSets", "Segmentation", self.split + ".txt"
            )
            file_list = tuple(open(file_list, "r"))
            file_list = [id_.rstrip() for id_ in file_list]
            self.files = file_list
        else:
            raise ValueError("Invalid split name: {}".format(self.split))

        self.image_ids = self.files

    def get_item(self, index):
        image_id = self.files[index]

        # * Load imgs
        image_path = osp.join(self.image_dir, image_id + ".jpg")
        if self.ret_img_file:
            image = image_path
        elif self.PIL_read:
            image = RGB2BGR(np.asarray(Image.open(image_path), dtype=np.uint8))
        else:
            image = cv2.imread(image_path, cv2.IMREAD_COLOR)

        # * Load Labels
        if self.split != 'test':
            label_path = osp.join(self.label_dir, image_id + ".png")
            label = np.asarray(Image.open(label_path), dtype=np.uint8)
        else:
            # If no label, then giving a all ignore ground truth
            img_size = imagesize.get(image)[::-1] if isinstance(image, str) else image.shape[:2]
            label = np.zeros(img_size, dtype=np.uint8) + self.ignore_label

        return image_id, image, label


class VOCAug(_VOCBase):
    """
    PASCAL VOC Aug Segmentation dataset
    """
    def __init__(self, root: str = "./contrib/datasets", year="2012", split: str = "train", PIL_read: bool=False,
                 ret_img_file: bool=False):
        """
        Args:
            root (str): The parent dir of VOC dataset
            year (str): "2012"
            split (str): "train"/"val"/"trainval"/"train_aug"/"trainval_aug"
            PIL_read (bool): If True, use PIL to read image, else use cv2
        """
        self.labels = []
        super(VOCAug, self).__init__(root, year, split, PIL_read, ret_img_file)

    def _set_files(self):
        self.root = osp.join(self.root, f"VOC{self.year}")
        self.image_dir = osp.join(self.root, "JPEGImages")
        self.label_dir = osp.join(self.root, "SegmentationClassAug")

        if self.split in ["train", "train_aug", "trainval", "trainval_aug", "val"]:
            file_list = osp.join(
                self.root, "ImageSets", "SegmentationAug", self.split + ".txt"
            )
            file_list = tuple(open(file_list, "r"))
            file_list = [id_.rstrip().split(" ") for id_ in file_list]
            self.files, self.labels = list(zip(*file_list))
        elif self.split == 'test':
            file_list = osp.join(
                self.root, "ImageSets", "SegmentationAug", self.split + ".txt"
            )
            file_list = tuple(open(file_list, "r"))
            self.files = [id_.rstrip() for id_ in file_list]
        else:
            raise ValueError("Invalid split name: {}".format(self.split))

        self.image_ids = [file.split("/")[-1].split(".")[0] for file in self.files]

    def get_item(self, index):
        # * Load image
        image_id = self.files[index].split("/")[-1].split(".")[0]
        image_path = osp.join(self.root, self.files[index][1:])
        if self.ret_img_file:
            image = image_path
        elif self.PIL_read:
            image = RGB2BGR(np.asarray(Image.open(image_path), dtype=np.uint8))
        else:
            image = cv2.imread(image_path, cv2.IMREAD_COLOR)

        # * Load Label
        if self.split != 'test':
            label_path = osp.join(self.root, self.labels[index][1:])
            label = np.asarray(Image.open(label_path), dtype=np.uint8)
        else:
            # If no label, then giving a all ignore ground truth
            img_size = imagesize.get(image)[::-1] if isinstance(image, str) else image.shape[:2]
            label = np.zeros(img_size, dtype=np.uint8) + self.ignore_label

        return image_id, image, label


class VOCAug2(VOCAug):
    """在标准VOCAug基础上:

    1. 返回cls_lb。
    2. 支持将lb替换为伪标签。
    3. 将输出打包为字典。
    4. 支持返回RGB img。
    """
    def __init__(self, root: str = "./contrib/datasets", year="2012", split: str = "train",
                 cls_labels_type: str='seg_cls_labels',
                 ps_mask_dir: str=None,
                 rgb_img: bool=False,
                 ret_img_file: bool=False):
        super().__init__(root, year, split, PIL_read=True, ret_img_file=ret_img_file)
        # * 参数检查与记录。
        assert cls_labels_type in ('seg_cls_labels', 'det_cls_labels', 'ignore_diff_cls_labels')
        self.cls_labels_type = cls_labels_type
        # * 读取图像级标签。
        with open(osp.join(self.root, 'third_party', f'{cls_labels_type}.pkl'), 'rb') as pkl_f:
            self.id2cls_labels = pickle.load(pkl_f)
        # * 记录伪真值目录。
        self.ps_mask_dir = ps_mask_dir
        # * 是否返回RGB图像。
        self.rgb_img = rgb_img

    def get_item(self, index: int) -> Dict:
        img_id, img, lb = super().get_item(index)

        if self.rgb_img:
            img = BGR2RGB(img).copy()

        out = Dict()
        out.img_id, out.img, out.lb = img_id, img, lb

        if self.split != 'test':
            out.cls_lb = self.id2cls_labels[img_id]
        else:
            out.cls_lb = np.zeros((self.class_num,), dtype=np.uint8)

        if self.ps_mask_dir is not None:
            out.lb = np.asarray(Image.open(osp.join(self.ps_mask_dir, f'{img_id}.png')), dtype=np.uint8)

        return out

    @classmethod
    def subset(cls,
               split_idx_num: tuple[int, int],
               root: str = "./contrib/datasets", year="2012", split: str = "train",
               cls_labels_type: str='seg_cls_labels',
               ps_mask_dir: str=None,
               rgb_img: bool=False):
        dt = cls(root, year, split, cls_labels_type, ps_mask_dir, rgb_img)

        split_idx, split_num = split_idx_num
        assert split_idx < split_num

        step = ceil(len(dt) / split_num)
        indexes = list(range(split_idx * step, min((split_idx + 1) * step, len(dt))))

        sub_dt = Subset(dt, indexes)
        return sub_dt

    @staticmethod
    def label_map2color_map(label_map: np.ndarray) -> np.ndarray:
        return label_map2color_map(label_map)


if __name__ == "__main__":
    from alchemy_cat.acplot.figure_wall import RectFigureWall, RowFigureWall
    import matplotlib.pyplot as plt

    voc = VOC()
    voc_aug = VOCAug(split="train_aug")

    voc_indexes = np.random.choice(len(voc), size=10, replace=False)
    voc_fig_wall = RectFigureWall([RowFigureWall([BGR2RGB(item[1]), label_map2color_map(item[2])], space_width=20)
                                   for item in voc[voc_indexes]],
                                  row_num=2, col_num=5, space_width=10)
    voc_fig_wall.plot(dpi=600)

    plt.show()

    vocaug_indexes = np.random.choice(len(voc_aug), size=10, replace=False)
    vocaug_fig_wall = RectFigureWall([RowFigureWall([BGR2RGB(item[1]), label_map2color_map(item[2])], space_width=40)
                                      for item in voc_aug[vocaug_indexes]],
                                     row_num=2, col_num=5, space_width=10)
    vocaug_fig_wall.plot(dpi=600)

    plt.show()